#
# Copyright (c) 2025 CESNET z.s.p.o.
#
# This file is a part of oarepo-requests (see http://github.com/oarepo/oarepo-requests).
#
# oarepo-requests is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.
#
"""OARepo notification builders base classes."""

from __future__ import annotations

from typing import TYPE_CHECKING, Protocol, override

from invenio_notifications.backends import EmailNotificationBackend
from invenio_notifications.models import Notification
from invenio_notifications.registry import EntityResolverRegistry
from invenio_notifications.services.builders import NotificationBuilder
from invenio_notifications.services.generators import ContextGenerator
from invenio_notifications.services.generators import (
    UserEmailBackend as InvenioUserEmailBackend,
)
from oarepo_runtime.typing import require_kwargs

from oarepo_requests.notifications.filters import (
    SystemUserRecipientFilter,
    UsersWithNoMailRecipientFilter,
)
from oarepo_requests.notifications.generators import ReferenceSavingEntityResolve

if TYPE_CHECKING:
    from invenio_notifications.services.filters import RecipientFilter
    from invenio_requests.records.api import Request


class NotificationBackendWithClassId(Protocol):
    """Protocol for notification backends with class id."""

    backend_id: str


class UserEmailBackend(NotificationBackendWithClassId, InvenioUserEmailBackend):
    """Email backend for user emails."""

    backend_id = EmailNotificationBackend.id


class RequestActionNotificationBuilder(NotificationBuilder):
    """Base class for request action notifications."""

    @classmethod
    @override
    @require_kwargs("request")
    def build(cls, request: Request) -> Notification:
        """Build notification with context."""
        return Notification(
            type=cls.type,
            context={
                "request": EntityResolverRegistry.reference_entity(request),
                "backend_ids": [backend.backend_id for backend in cls.recipient_backends],
            },
        )

    context: tuple[ContextGenerator, ...] = (
        ReferenceSavingEntityResolve(key="request"),
        ReferenceSavingEntityResolve(key="request.topic"),
        ReferenceSavingEntityResolve(key="request.created_by"),
        ReferenceSavingEntityResolve(key="request.receiver"),
    )

    recipient_backends: tuple[NotificationBackendWithClassId, ...] = (UserEmailBackend(),)

    recipient_filters: tuple[RecipientFilter, ...] = (
        UsersWithNoMailRecipientFilter(),
        SystemUserRecipientFilter(),
    )



#
# Copyright (C) 2025 CESNET z.s.p.o.
#
# oarepo-requests is free software; you can redistribute it and/or
# modify it under the terms of the MIT License; see LICENSE file for more
# details.
#
"""Notification context generators in oarepo-requests."""

from __future__ import annotations

from typing import TYPE_CHECKING, Any, override

from invenio_access.permissions import system_identity
from invenio_notifications.services.generators import EntityResolve
from invenio_records.dictutils import dict_lookup
from invenio_requests.records import Request

if TYPE_CHECKING:
    from invenio_notifications.models import Notification


class NotificationCtxWithReference(dict):
    """Context dict that keeps up separate dict with unresolved references."""

    def __init__(
        self,
        reference_dict: dict[str, str],
        reference_key: str,
        ctx: dict[str, Any] | NotificationCtxWithReference,
    ):
        """Initialize the context."""
        super().__init__(ctx)
        self.references: dict[str, dict[str, str]] = (
            ctx.references if isinstance(ctx, NotificationCtxWithReference) else {}
        )
        self.references[reference_key] = reference_dict

    def get_reference_type(self, key: str) -> str:
        """Get the type of the reference stored under the given key."""
        return next(iter(self.references[key].keys()))


class ReferenceSavingEntityResolve(EntityResolve):
    """Entity resolver that saves the reference in the context."""

    @override
    def __call__(self, notification: Notification):
        entity_ref = dict_lookup(notification.context, self.key)
        if entity_ref is None:
            return notification
        notification = super().__call__(notification)
        notification.context = NotificationCtxWithReference(entity_ref, self.key, notification.context)
        return notification


class RequestEntityResolve(ReferenceSavingEntityResolve):
    """Entity resolver that adds the correct title if it is missing."""

    @override
    def __call__(self, notification: Notification):
        notification = super().__call__(notification)
        request_dict = notification.context["request"]
        if request_dict.get("title"):
            return request_dict

        request = Request.get_record(request_dict["id"])
        if hasattr(request.type, "stateful_name"):
            # If the request type has a stateful name, use it
            # note: do not have better identity here, so using system_identity
            # as a fallback
            request_dict["title"] = request.type.stateful_name(system_identity, topic=None)  # type: ignore[reportAttributeAccessIssue]
        return notification



#
# Copyright (C) 2024 CESNET z.s.p.o.
#
# oarepo-requests is free software; you can redistribute it and/or
# modify it under the terms of the MIT License; see LICENSE file for more
# details.
#
"""Command Line Interface (CLI) commands for OARepo requests.

This module provides CLI commands for managing OARepo requests,
including request escalation functionality.
"""

from __future__ import annotations

from oarepo_runtime.cli import oarepo


@oarepo.group(name="requests")
def oarepo_requests() -> None:
    """OARepo requests group command."""


# TODO: escalate_requests



#
# Copyright (c) 2025 CESNET z.s.p.o.
#
# This file is a part of oarepo-requests (see https://github.com/oarepo/oarepo-communities).
#
# oarepo-communities is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.
#
"""Notifications entity resolvers."""

from __future__ import annotations

import json
from typing import Any, cast, override

from invenio_access.permissions import system_user_id
from invenio_notifications.registry import EntityResolverRegistry
from invenio_records_resources.references import EntityResolver
from invenio_records_resources.references.entity_resolvers import (
    EntityProxy,
    ServiceResultProxy,
    ServiceResultResolver,
)
from invenio_records_resources.services.errors import PermissionDeniedError
from invenio_users_resources.entity_resolvers import UserProxy
from oarepo_workflows.resolvers.multiple_entities import (
    MultipleEntitiesProxy,
    MultipleEntitiesResolver,
)


class UserNotificationProxy(ServiceResultProxy):
    """Entity proxy for user notifications."""

    system_record = UserProxy.system_record
    ghost_record = UserProxy.ghost_record

    def _resolve(self) -> dict[str, Any]:
        """Resolve the User from the proxy's reference dict, or system_identity."""
        user_id = self._parse_ref_dict_id()
        if user_id == system_user_id:  # system_user_id is a string: "system"
            return self.system_record()  # type: ignore[no-any-return]
        try:
            return super()._resolve()  # type: ignore[no-any-return]
        except PermissionDeniedError:
            return self.ghost_record({"id": user_id})  # type: ignore[no-any-return]


class MultipleEntitiesNotificationProxy(EntityProxy):
    """Entity proxy for email addresses."""

    # not used in notifications but required for inheritance
    get_needs = MultipleEntitiesProxy.get_needs
    pick_resolved_fields = MultipleEntitiesProxy.pick_resolved_fields

    @override
    def _resolve(self) -> list[dict[str, dict[str, Any]]]:  # type: ignore[reportIncompatibleMethodOverride]
        """Resolve the User from the proxy's reference dict, or system_identity."""
        entity_refs = json.loads(self._parse_ref_dict_id())

        fields: list[dict[str, dict[str, Any]]] = []
        for entity_ref in entity_refs:
            type_ = next(iter(entity_ref.keys()))
            fields.append(
                {
                    type_: cast(
                        "dict[str, Any]",
                        EntityResolverRegistry.resolve_entity(entity_ref),
                    )
                }
            )

        return fields


class MultipleEntitiesNotificationResolver(EntityResolver):
    """Resolver for user notifications."""

    type_id = MultipleEntitiesResolver.type_id
    matches_reference_dict = MultipleEntitiesResolver.matches_reference_dict
    matches_entity = MultipleEntitiesResolver.matches_entity
    _reference_entity = MultipleEntitiesResolver._reference_entity  # noqa SLF001

    def __init__(self) -> None:
        """Initialize the resolver."""
        super().__init__("multiple")

    def _get_entity_proxy(self, ref_dict: dict[str, str]) -> EntityProxy:
        """Return a UserProxy for the given reference dict."""
        return MultipleEntitiesNotificationProxy(self, ref_dict)


def requests_resolver() -> ServiceResultResolver:
    """Return community role notification resolver."""
    return ServiceResultResolver(service_id="requests", type_key="request")


def request_events_resolver() -> ServiceResultResolver:
    """Return community role notification resolver."""
    return ServiceResultResolver(service_id="request_events", type_key="request_event")


def user_resolver() -> ServiceResultResolver:
    """Return community role notification resolver."""
    return ServiceResultResolver(service_id="users", type_key="user", proxy_cls=UserNotificationProxy)


def multiple_entities_resolver() -> MultipleEntitiesNotificationResolver:
    """Return community role notification resolver."""
    return MultipleEntitiesNotificationResolver()


requests_resolver.type_key = "request"  # type: ignore[attr-defined]
request_events_resolver.type_key = "request_event"  # type: ignore[attr-defined]
user_resolver.type_key = "user"  # type: ignore[attr-defined]
multiple_entities_resolver.type_key = MultipleEntitiesNotificationResolver.type_id  # type: ignore[attr-defined]



#
# Copyright (C) 2024 CESNET z.s.p.o.
#
# oarepo-requests is free software; you can redistribute it and/or
# modify it under the terms of the MIT License; see LICENSE file for more
# details.
#
"""Request needs."""

from __future__ import annotations

from invenio_access.permissions import SystemRoleNeed

request_active = SystemRoleNeed("request")
"""Need that is added to identity whenever a request is being handled (inside the 'accept' action)."""



#
# Copyright (C) 2024 CESNET z.s.p.o.
#
# oarepo-requests is free software; you can redistribute it and/or
# modify it under the terms of the MIT License; see LICENSE file for more
# details.
#
"""OARepo extensions to invenio requests resource."""

from __future__ import annotations

from flask import g
from flask_resources import resource_requestctx, response_handler, route
from invenio_records_resources.resources.errors import ErrorHandlersMixin
from invenio_records_resources.resources.records.resource import (
    request_data,
    request_extra_args,
    request_headers,
    request_search_args,
    request_view_args,
)
from invenio_requests.resources import RequestsResource

from oarepo_requests.utils import (
    resolve_reference_dict,
    string_to_reference,
)


class OARepoRequestsResource(RequestsResource, ErrorHandlersMixin):
    """OARepo extensions to invenio requests resource."""

    def create_url_rules(self) -> list[dict]:
        """Create the URL rules for the record resource."""

        def p(route: str) -> str:
            """Prefix a route with the URL prefix."""
            return f"{self.config.url_prefix}{route}"

        routes = self.config.routes

        return [
            *super().create_url_rules(),
            route("POST", p(routes["list"]), self.create),
            route("POST", p(routes["create"]), self.create_via_url),
            route("GET", p(routes["list-applicable"]), self.applicable_request_types),
        ]

    @request_extra_args
    @request_headers
    @request_view_args
    @request_data
    @response_handler()
    def create(self) -> tuple[dict, int]:
        """Create a new request based on a request type.

        The data is in the form of:
            .. code-block:: json
            {
                "request_type": "request_type",
                "topic": {
                    "type": "pid",
                    "value": "value"
                },
                ...payload
            }
        """
        request_type_id = resource_requestctx.data.pop("request_type", None)
        topic = resource_requestctx.data.pop("topic", None)
        if isinstance(topic, str):
            topic = string_to_reference(topic)
        topic = resolve_reference_dict(topic)

        items = self.service.create(
            identity=g.identity,
            data=resource_requestctx.data,
            request_type=request_type_id,
            topic=topic,
            expand=resource_requestctx.args.get("expand", False),
        )

        return items.to_dict(), 201

    @request_extra_args
    @request_view_args
    @request_headers
    @request_data
    @response_handler()
    def create_via_url(self) -> tuple[dict, int]:
        """Create a new request based on a request type and topic from view_args arguments."""
        request_type_id = resource_requestctx.view_args["request_type"]
        topic = resolve_reference_dict(string_to_reference(resource_requestctx.view_args["topic"]))

        items = self.service.create(
            identity=g.identity,
            data=resource_requestctx.data,
            request_type=request_type_id,
            topic=topic,
            expand=resource_requestctx.args.get("expand", False),
        )

        return items.to_dict(), 201

    @request_search_args
    @request_view_args
    @response_handler(many=True)
    def applicable_request_types(self) -> tuple[dict, int]:
        """List request types."""
        hits = self.service.applicable_request_types(
            identity=g.identity,
            topic=resolve_reference_dict(resource_requestctx.args["topic"]),
        )
        return hits.to_dict(), 200



#
# Copyright (C) 2024 CESNET z.s.p.o.
#
# oarepo-requests is free software; you can redistribute it and/or
# modify it under the terms of the MIT License; see LICENSE file for more
# details.
#
"""OARepo-Requests extension."""

from __future__ import annotations

from typing import TYPE_CHECKING, cast

from deepmerge import conservative_merger
from invenio_base.utils import obj_or_import_string
from invenio_requests.customizations import EventType
from invenio_requests.proxies import (
    current_event_type_registry,
    current_request_type_registry,
)

if TYPE_CHECKING:
    from flask import Flask
    from flask_principal import Identity
    from invenio_records_resources.records.api import Record
    from invenio_requests.customizations import RequestType

    from oarepo_requests.actions.components import RequestActionComponent


class _OARepoRequestsState:
    def __init__(self, app: Flask):
        """Initialize state.

        :param app: An instance of :class:`~flask.app.Flask`.
        :param entry_point_group_mappings:
            The entrypoint group name to load mappings.
        :param entry_point_group_templates:
            The entrypoint group name to load templates.
        """
        self.app = app

    def init_app(self, app: Flask) -> None:
        """Flask application initialization."""
        self.app = app
        app.extensions["oarepo-requests"] = self

    @property
    def allowed_receiver_ref_types(self) -> list[str]:
        """Return a list of allowed receiver entity reference types.

        This value is taken from the configuration key REQUESTS_ALLOWED_RECEIVERS.
        """
        return cast("list[str]", self.app.config.get("REQUESTS_ALLOWED_RECEIVERS", []))

    def action_components(self) -> list[type[RequestActionComponent]]:
        """Return components for the given action."""
        return cast(
            "list[type[RequestActionComponent]]",
            self.app.config["REQUESTS_ACTION_COMPONENTS"],
        )

    def default_request_receiver(
        self,
        identity: Identity,
        request_type: RequestType,
        record: Record,
        creator: dict[str, str] | Identity,
        data: dict,
    ) -> dict[str, str] | None:
        """Return the default receiver for the request.

        Gets the default receiver for the request based on the request type, record and data.
        It is used when the receiver is not explicitly set when creating a request. It does so
        by taking a function from the configuration under the key OAREPO_REQUESTS_DEFAULT_RECEIVER
        and calling it with the given parameters.

        :param identity: Identity of the user creating the request.
        :param request_type: Type of the request.
        :param record: Record the request is about.
        :param creator: Creator of the request.
        :param data: Payload of the request.
        """
        return obj_or_import_string(self.app.config["OAREPO_REQUESTS_DEFAULT_RECEIVER"])(  # type: ignore[no-any-return]
            identity=identity,
            request_type=request_type,
            record=record,
            creator=creator,
            data=data,
        )


class OARepoRequests:
    """OARepo-Requests extension."""

    def __init__(self, app: Flask | None = None) -> None:
        """Extension initialization."""
        if app:
            self.init_app(app)

    def init_app(self, app: Flask) -> None:
        """Flask application initialization.

        :param app: An instance of :class:`~flask.app.Flask`.
        """
        self.init_config(app)
        state = _OARepoRequestsState(app)
        self._state = app.extensions["oarepo-requests"] = state

    def __getattr__(self, name: str):
        """Proxy to state object."""
        return getattr(self._state, name, None)

    def init_config(self, app: Flask) -> None:
        """Initialize configuration."""
        from . import config

        app.config.setdefault("OAREPO_REQUESTS_DEFAULT_RECEIVER", None)
        app.config.setdefault("REQUESTS_ALLOWED_RECEIVERS", []).extend(config.REQUESTS_ALLOWED_RECEIVERS)

        app.config.setdefault("PUBLISH_REQUEST_TYPES", config.PUBLISH_REQUEST_TYPES)

        # do not overwrite user's stuff
        app_default_workflow_events = app.config.setdefault("DEFAULT_WORKFLOW_EVENTS", {})
        for k, v in config.DEFAULT_WORKFLOW_EVENTS.items():
            if k not in app_default_workflow_events:
                app_default_workflow_events[k] = v

        # let the user override the action components
        app.config.setdefault("REQUESTS_ACTION_COMPONENTS", []).extend(config.REQUESTS_ACTION_COMPONENTS)

        app_notification_recipient_resolvers = app.config.setdefault("NOTIFICATION_RECIPIENTS_RESOLVERS", {})
        app.config["NOTIFICATION_RECIPIENTS_RESOLVERS"] = conservative_merger.merge(
            app_notification_recipient_resolvers,
            config.NOTIFICATION_RECIPIENTS_RESOLVERS,
        )

        app_notification_builders = app.config.setdefault("NOTIFICATIONS_BUILDERS", {})
        app_notification_backends = app.config.setdefault("NOTIFICATIONS_BACKENDS", {})

        app.config["NOTIFICATIONS_BUILDERS"] = conservative_merger.merge(
            app_notification_builders, config.NOTIFICATIONS_BUILDERS
        )
        app.config["NOTIFICATIONS_BACKENDS"] = conservative_merger.merge(
            app_notification_backends, config.NOTIFICATIONS_BACKENDS
        )


def api_finalize_app(app: Flask) -> None:
    """Finalize app."""
    finalize_app(app)


def finalize_app(app: Flask) -> None:
    """Finalize app."""
    # TODO: temporary before invenio fix
    to_remove = [t for t in current_request_type_registry if isinstance(t, EventType)]
    for type_ in to_remove:
        current_request_type_registry._registered_types.pop(type_.type_id)  # noqa SLF001 # type: ignore[reportAttributeAccessIssue]
        current_event_type_registry.register_type(type_)

    ext = app.extensions["oarepo-requests"]
    ext.notification_recipients_resolvers_registry = app.config["NOTIFICATION_RECIPIENTS_RESOLVERS"]

    invenio_notifications = app.extensions["invenio-notifications"]
    invenio_notifications.init_manager(app)



#
# Copyright (c) 2025 CESNET z.s.p.o.
#
# This file is a part of oarepo-requests (see https://github.com/oarepo/oarepo-requests).
#
# oarepo-requests is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.
#

"""Components for request service functionality.

This module provides service components for handling request operations
and conditions.
"""

from __future__ import annotations



#
# Copyright (c) 2025 CESNET z.s.p.o.
#
# This file is a part of oarepo-model (see http://github.com/oarepo/oarepo-model).
#
# oarepo-model is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.
#
"""Module to add requests-relevant metadata mappings on record."""

from __future__ import annotations

from typing import TYPE_CHECKING, Any, override

from oarepo_model.customizations import Customization, PatchJSONFile
from oarepo_model.presets import Preset

if TYPE_CHECKING:
    from collections.abc import Generator

    from oarepo_model.builder import InvenioModelBuilder
    from oarepo_model.model import InvenioModel


class RequestsMetadataMappingPreset(Preset):
    """Preset for applying requests-related record mapping changes."""

    modifies = ("record-mapping",)

    @override
    def apply(
        self,
        builder: InvenioModelBuilder,
        model: InvenioModel,
        dependencies: dict[str, Any],
    ) -> Generator[Customization]:
        yield PatchJSONFile(
            "record-mapping",
            {
                "mappings": {
                    "properties": {
                        "metadata": {
                            "properties": {
                                "version": {"type": "keyword"},
                            },
                        },
                    },
                },
            },
        )



#
# Copyright (c) 2025 CESNET z.s.p.o.
#
# This file is a part of oarepo-requests (see http://github.com/oarepo/oarepo-requests).
#
# oarepo-requests is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.
#
"""OARepo requests model presets.

This module provides a collection of presets for configuring requests functionality
in OARepo models. The presets include components for API blueprints, service configurations,
entity resolvers, metadata mappings, and application finalization.

The requests_presets list contains all available presets in the correct order for
application during model building.
"""

from __future__ import annotations

from oarepo_requests.model.presets.requests.records.metadata_mapping import (
    RequestsMetadataMappingPreset,
)
from oarepo_requests.model.presets.requests.register_resolvers import (
    RegisterResolversPreset,
)
from oarepo_requests.model.presets.requests.services.records.results import (
    RequestsRecordItemPreset,
)
from oarepo_requests.model.presets.requests.services.records.service_config import (
    RequestsServiceConfigPreset,
)

# Collection of all request-related presets in proper initialization order
requests_preset = [
    RequestsMetadataMappingPreset,  # Configure metadata mapping for requests
    RequestsServiceConfigPreset,  # Configure service with request components
    RequestsRecordItemPreset,  # Configure record item results
    RegisterResolversPreset,  # Final application setup
]



#
# Copyright (C) 2024 CESNET z.s.p.o.
#
# oarepo-requests is free software; you can redistribute it and/or
# modify it under the terms of the MIT License; see LICENSE file for more
# details.
#
"""Default configuration of oarepo-requests."""

from __future__ import annotations

from invenio_app_rdm.config import NOTIFICATIONS_BUILDERS as RDM_NOTIFICATIONS_BUILDERS
from invenio_notifications.backends.email import EmailNotificationBackend
from invenio_requests.customizations import CommentEventType, LogEventType
from invenio_requests.services.permissions import (
    PermissionPolicy as InvenioRequestsPermissionPolicy,
)
from invenio_users_resources.notifications.generators import UserRecipient
from oarepo_workflows.requests.events import WorkflowEvent

from oarepo_requests.actions.components import (
    AutoAcceptComponent,
    RequestActionComponent,
    WorkflowTransitionComponent,
)
from oarepo_requests.notifications.builders.comment import (
    CommentRequestEventCreateNotificationBuilder,
)
from oarepo_requests.notifications.builders.delete_published_record import (
    DeletePublishedRecordRequestAcceptNotificationBuilder,
    DeletePublishedRecordRequestDeclineNotificationBuilder,
    DeletePublishedRecordRequestSubmitNotificationBuilder,
)
from oarepo_requests.notifications.builders.publish import (
    PublishDraftRequestAcceptNotificationBuilder,
    PublishDraftRequestDeclineNotificationBuilder,
    PublishDraftRequestSubmitNotificationBuilder,
)
from oarepo_requests.notifications.generators import MultipleRecipients

# check individual receivers in multiple?
REQUESTS_ALLOWED_RECEIVERS = ["user", "group", "auto_approve"]

DEFAULT_WORKFLOW_EVENTS = {
    CommentEventType.type_id: WorkflowEvent(submitters=InvenioRequestsPermissionPolicy.can_create_comment),
    LogEventType.type_id: WorkflowEvent(submitters=InvenioRequestsPermissionPolicy.can_create_comment),
}

REQUESTS_ACTION_COMPONENTS: tuple[type[RequestActionComponent], ...] = (
    WorkflowTransitionComponent,
    AutoAcceptComponent,
)

# TODO: possibly not used outside ui
PUBLISH_REQUEST_TYPES = ["publish_draft", "publish_new_version"]


NOTIFICATION_RECIPIENTS_RESOLVERS = {
    "user": lambda key, notification: UserRecipient(key),  # noqa ARG005
    "multiple": lambda key, notification: MultipleRecipients(key),  # noqa ARG005
}

NOTIFICATIONS_BACKENDS = {
    EmailNotificationBackend.id: EmailNotificationBackend(),
}

NOTIFICATIONS_BUILDERS = {
    **RDM_NOTIFICATIONS_BUILDERS,
    DeletePublishedRecordRequestSubmitNotificationBuilder.type: DeletePublishedRecordRequestSubmitNotificationBuilder,
    DeletePublishedRecordRequestAcceptNotificationBuilder.type: DeletePublishedRecordRequestAcceptNotificationBuilder,
    DeletePublishedRecordRequestDeclineNotificationBuilder.type: DeletePublishedRecordRequestDeclineNotificationBuilder,
    PublishDraftRequestSubmitNotificationBuilder.type: PublishDraftRequestSubmitNotificationBuilder,
    PublishDraftRequestAcceptNotificationBuilder.type: PublishDraftRequestAcceptNotificationBuilder,
    PublishDraftRequestDeclineNotificationBuilder.type: PublishDraftRequestDeclineNotificationBuilder,
    CommentRequestEventCreateNotificationBuilder.type: CommentRequestEventCreateNotificationBuilder,
}



#
# Copyright (C) 2025 CESNET z.s.p.o.
#
# oarepo-requests is free software; you can redistribute it and/or
# modify it under the terms of the MIT License; see LICENSE file for more
# details.
#
#
# This file is part of Invenio.
# Copyright (C) 2016-2018 CERN.
#
# Invenio is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Create oarepo requests branch."""

from __future__ import annotations

# revision identifiers, used by Alembic.
revision = "985368490c5b"
down_revision = None
branch_labels = ("oarepo_requests",)
depends_on = None


def upgrade() -> None:
    """Upgrade database."""
    # ### commands auto generated by Alembic - please adjust! ###
    # ### end Alembic commands ###


def downgrade() -> None:
    """Downgrade database."""
    # ### commands auto generated by Alembic - please adjust! ###
    # ### end Alembic commands ###



#
# Copyright (C) 2024 CESNET z.s.p.o.
#
# oarepo-requests is free software; you can redistribute it and/or
# modify it under the terms of the MIT License; see LICENSE file for more
# details.
#
"""Results components for requests."""

from __future__ import annotations

from typing import TYPE_CHECKING, Any, cast

from invenio_records_resources.records import Record
from invenio_records_resources.services.records.schema import ServiceSchemaWrapper
from invenio_requests.services.results import EntityResolverExpandableField
from oarepo_runtime.services.results import RecordList

from oarepo_requests.services.schema import (
    RequestTypeSchema,
    request_type_identity_ctx,
    request_type_record_ctx,
)
from oarepo_requests.utils import string_to_reference

if TYPE_CHECKING:
    from collections.abc import Iterator

    from flask_principal import Identity
    from invenio_records_resources.records.api import Record
    from invenio_records_resources.services.records.service import RecordService
    from invenio_requests.customizations import RequestType


class StringEntityResolverExpandableField(EntityResolverExpandableField):
    """Expandable entity resolver field.

    It will use the Entity resolver registry to retrieve the service to
    use to fetch records and the fields to return when serializing
    the referenced record.
    """

    def get_value_service(self, value: str) -> tuple[str, RecordService]:  # type: ignore[override]
        """Return the value and the service via entity resolvers."""
        ref = string_to_reference(value)
        v, service = super().get_value_service(ref)
        return v, service


class RequestTypesListDict(dict):
    """List of request types dictionary with additional topic."""

    topic: Record | None = None


class RequestTypesList(RecordList):
    """An in-memory list of request types compatible with opensearch record list."""

    def __init__(
        self,
        *args: Any,
        record: Record | None = None,
        schema: ServiceSchemaWrapper | None = None,
        **kwargs: Any,
    ) -> None:
        """Initialize the list of request types."""
        self._record = record
        super().__init__(*args, **kwargs)
        self._schema = schema or ServiceSchemaWrapper(self._service, RequestTypeSchema)

    def to_dict(self) -> dict:
        """Return result as a dictionary."""
        hits = list(self.hits)
        res = RequestTypesListDict(
            hits={
                "hits": hits,
                "total": self.total,
            }
        )
        if self._links_tpl:
            res["links"] = self._links_tpl.expand(self._identity, None)
        res.topic = self._record
        return res

    @property
    def hits(self) -> Iterator[dict]:
        """Iterator over the hits."""
        for hit in self._results:
            # Project the record
            tok_identity = request_type_identity_ctx.set(self._identity)
            tok_record = request_type_record_ctx.set(self._record)
            try:
                # identity in context is hardcoded in ServiceSchemaWrapper
                # which we have to use if we want to subclass RecordList
                projection = self._schema.dump(hit, context={"identity": self._identity})
            finally:
                # Reset contextvars to previous values to avoid leaking state
                request_type_identity_ctx.reset(tok_identity)
                request_type_record_ctx.reset(tok_record)

            if self._links_item_tpl:
                projection["links"] = self._links_item_tpl.expand(self._identity, hit)
            yield projection

    @property
    def total(self) -> int:
        """Total number of hits."""
        return len(self._results)


def serialize_request_types(
    request_types: dict[str, RequestType], identity: Identity, record: Record
) -> list[dict[str, Any]]:
    """Serialize request types.

    :param request_types: Request types to serialize.
    :param identity: Identity of the user.
    :param record: Record for which the request types are serialized.
    :return: List of serialized request types.
    """
    # contextvars approach from gpt
    tok_identity = request_type_identity_ctx.set(identity)
    tok_record = request_type_record_ctx.set(record)
    try:
        return [
            cast("dict[str, Any]", RequestTypeSchema().dump(request_type)) for request_type in request_types.values()
        ]
    finally:
        # Reset contextvars to previous values to avoid leaking state
        request_type_identity_ctx.reset(tok_identity)
        request_type_record_ctx.reset(tok_record)



#
# Copyright (C) 2024 CESNET z.s.p.o.
#
# oarepo-requests is free software; you can redistribute it and/or
# modify it under the terms of the MIT License; see LICENSE file for more
# details.
#
"""Permission generators."""

from __future__ import annotations

from .active import RequestActive
from .conditional import (
    IfNoEditDraft,
    IfNoNewVersionDraft,
    IfRequestedBy,
)

__all__ = (
    "IfNoEditDraft",
    "IfNoNewVersionDraft",
    "IfRequestedBy",
    "RequestActive",
)



#
# Copyright (C) 2024 CESNET z.s.p.o.
#
# oarepo-requests is free software; you can redistribute it and/or
# modify it under the terms of the MIT License; see LICENSE file for more
# details.
#
"""Proxy objects for accessing the current application's requests service and resource."""

from __future__ import annotations

from typing import TYPE_CHECKING

from flask import current_app
from invenio_requests.proxies import (
    current_requests_service as current_invenio_requests_service,
)
from werkzeug.local import LocalProxy

if TYPE_CHECKING:
    from oarepo_requests.ext import _OARepoRequestsState
    from oarepo_requests.resources.oarepo.resource import OARepoRequestsResource
    from oarepo_requests.services.oarepo.service import OARepoRequestsService

current_oarepo_requests: _OARepoRequestsState = LocalProxy(  # type: ignore[assignment]
    lambda: current_app.extensions["oarepo-requests"]
)
current_requests_service: OARepoRequestsService = current_invenio_requests_service  # type: ignore[assignment]
current_oarepo_requests_resource: OARepoRequestsResource = LocalProxy(  # type: ignore[assignment]
    lambda: current_app.extensions["oarepo-requests"].requests_resource
)

current_notification_recipient_generators_registry = LocalProxy(  # type: ignore[assignment]
    lambda: current_app.extensions["oarepo-requests"].notification_recipients_resolvers_registry
)



#
# Copyright (C) 2024 CESNET z.s.p.o.
#
# oarepo-requests is free software; you can redistribute it and/or
# modify it under the terms of the MIT License; see LICENSE file for more
# details.
#
"""Conditional generators for needs based on request type, event type, and requester role."""

from __future__ import annotations

from typing import TYPE_CHECKING, Any, override

from flask_principal import Identity
from invenio_records_permissions.generators import ConditionalGenerator, Generator
from invenio_records_resources.references.entity_resolvers import EntityProxy
from invenio_requests.resolvers.registry import ResolverRegistry
from oarepo_runtime.proxies import current_runtime
from oarepo_runtime.typing import require_kwargs
from oarepo_workflows.requests import RecipientGeneratorMixin
from sqlalchemy.exc import NoResultFound

if TYPE_CHECKING:
    from collections.abc import Mapping

    from invenio_drafts_resources.records import Record as DraftRecord
    from invenio_records_resources.records import Record
    from invenio_requests.customizations import RequestType
    from opensearch_dsl.query import Query


class IfRequestedBy(RecipientGeneratorMixin, ConditionalGenerator):
    """Conditional generator that generates needs when a request is made by a given requester role."""

    def __init__(
        self,
        requesters: list[Generator] | tuple[Generator] | Generator,
        then_: list[Generator],
        else_: list[Generator],
    ) -> None:
        """Initialize the generator."""
        super().__init__(then_, else_)
        if not isinstance(requesters, (list, tuple)):
            requesters = [requesters]
        self.requesters = requesters

    def _condition(
        self,
        *,
        request_type: RequestType,
        creator: Identity | EntityProxy | Any,
        **kwargs: Any,
    ) -> bool:
        """Condition to choose generators set."""
        # get needs
        if isinstance(creator, Identity):
            needs = creator.provides
        else:
            if not isinstance(creator, EntityProxy):
                creator = ResolverRegistry.reference_entity(creator)
            needs = creator.get_needs()

        for condition in self.requesters:
            condition_needs = set(condition.needs(request_type=request_type, creator=creator, **kwargs))
            condition_excludes = set(condition.excludes(request_type=request_type, creator=creator, **kwargs))

            if not condition_needs.intersection(needs):
                continue
            if condition_excludes and condition_excludes.intersection(needs):
                continue
            return True
        return False

    def reference_receivers(
        self,
        record: Record | None = None,
        request_type: RequestType | None = None,
        **context: Any,
    ) -> list[Mapping[str, str]]:  # pragma: no cover
        """Return the reference receiver(s) of the request.

        This call requires the context to contain at least "record" and "request_type"

        Must return a list of dictionary serialization of the receivers.

        Might return empty list or None to indicate that the generator does not
        provide any receivers.
        """
        ret = []
        for gen in self._generators(record=record, request_type=request_type, **context):
            if isinstance(gen, RecipientGeneratorMixin):
                ret.extend(gen.reference_receivers(record=record, request_type=request_type, **context))
        return ret

    def query_filter(self, **context: Any) -> Query:
        """Search filters."""
        raise NotImplementedError("Please use IfRequestedBy only in recipients, not elsewhere.")


class IfNoNewVersionDraft(ConditionalGenerator):
    """Generator that checks if the record has no new version draft."""

    def __init__(self, then_: list[Generator], else_: list[Generator] | None = None) -> None:
        """Initialize the generator."""
        else_ = [] if else_ is None else else_
        super().__init__(then_, else_=else_)

    @override
    @require_kwargs("record")
    def _condition(self, record: DraftRecord, **kwargs: Any) -> bool:
        if hasattr(record, "is_draft"):
            is_draft = record.is_draft
        else:
            return False
        if hasattr(record, "versions"):
            next_draft_id = record.versions.next_draft_id
        else:
            return False
        return not is_draft and not next_draft_id


class IfNoEditDraft(ConditionalGenerator):
    """Generator that checks if the record has no edit draft."""

    def __init__(self, then_: list[Generator], else_: list[Generator] | None = None) -> None:
        """Initialize the generator."""
        else_ = [] if else_ is None else else_
        super().__init__(then_, else_=else_)

    @override
    @require_kwargs("record")
    def _condition(self, record: DraftRecord, **kwargs: Any) -> bool:
        if getattr(record, "is_draft", False):
            return False
        records_service = current_runtime.get_record_service_for_record(record)
        try:
            records_service.config.draft_cls.pid.resolve(record["id"])  # by edit - has the same id as parent record
            # I'm not sure what open unpublished means
        except NoResultFound:
            return True
        else:
            return False



#
# Copyright (c) 2025 CESNET z.s.p.o.
#
# This file is a part of oarepo-model (see http://github.com/oarepo/oarepo-model).
#
# oarepo-model is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.
#
"""Module providing preset for configuring requests result classes."""

from __future__ import annotations

from typing import TYPE_CHECKING, Any, override

from oarepo_model.customizations import (
    AddToList,
    Customization,
)
from oarepo_model.presets import Preset

from oarepo_requests.services.components.results import (
    RequestsComponent,
    RequestTypesComponent,
)

if TYPE_CHECKING:
    from collections.abc import Generator

    from oarepo_model.builder import InvenioModelBuilder
    from oarepo_model.model import InvenioModel


class RequestsRecordItemPreset(Preset):
    """Preset for configuring requests result classes."""

    modifies = ("record_result_item_components",)

    @override
    def apply(
        self,
        builder: InvenioModelBuilder,
        model: InvenioModel,
        dependencies: dict[str, Any],
    ) -> Generator[Customization]:
        yield AddToList("record_result_item_components", RequestsComponent)
        yield AddToList("record_result_item_components", RequestTypesComponent)



#
# Copyright (C) 2024 CESNET z.s.p.o.
#
# oarepo-requests is free software; you can redistribute it and/or
# modify it under the terms of the MIT License; see LICENSE file for more
# details.
#
"""Request action components."""

from __future__ import annotations

import logging
from typing import TYPE_CHECKING, Any, override

from invenio_requests.customizations import RequestActions
from invenio_requests.errors import CannotExecuteActionError

if TYPE_CHECKING:
    from flask_principal import Identity
    from invenio_db.uow import UnitOfWork

    from .generic import OARepoGenericActionMixin

log = logging.getLogger(__name__)


class RequestActionComponent:
    """Abstract request action component.

    Implementation warning: calling actions reloading and recommiting request or topic (eg. topic publish) within
    components may cause incorrect behavior due to the fact that changes on the new (reloaded) instance do not
    propagate to commit ops in the stack below.
    """

    def create(
        self,
        identity: Identity,
        action: OARepoGenericActionMixin,
        uow: UnitOfWork,
        *args: Any,
        **kwargs: Any,
    ) -> None:
        """Apply the component on the create action.

        :param identity: Identity of the user.
        :param request_type: Request type.
        :param action: Action being executed.
        :param topic: Topic of the request.
        :param uow: Unit of work.
        :param args: Additional arguments.
        :param kwargs: Additional keyword arguments.
        """

    def submit(
        self,
        identity: Identity,
        action: OARepoGenericActionMixin,
        uow: UnitOfWork,
        *args: Any,
        **kwargs: Any,
    ) -> None:
        """Apply the component on the submit action.

        :param identity: Identity of the user.
        :param request_type: Request type.
        :param action: Action being executed.
        :param topic: Topic of the request.
        :param uow: Unit of work.
        :param args: Additional arguments.
        :param kwargs: Additional keyword arguments.
        """

    def accept(
        self,
        identity: Identity,
        action: OARepoGenericActionMixin,
        uow: UnitOfWork,
        *args: Any,
        **kwargs: Any,
    ) -> None:
        """Apply the component on the accept action.

        :param identity: Identity of the user.
        :param request_type: Request type.
        :param action: Action being executed.
        :param topic: Topic of the request.
        :param uow: Unit of work.
        :param args: Additional arguments.
        :param kwargs: Additional keyword arguments.
        """

    def decline(
        self,
        identity: Identity,
        action: OARepoGenericActionMixin,
        uow: UnitOfWork,
        *args: Any,
        **kwargs: Any,
    ) -> None:
        """Apply the component on the decline action.

        :param identity: Identity of the user.
        :param request_type: Request type.
        :param action: Action being executed.
        :param topic: Topic of the request.
        :param uow: Unit of work.
        :param args: Additional arguments.
        :param kwargs: Additional keyword arguments.
        """

    def cancel(
        self,
        identity: Identity,
        action: OARepoGenericActionMixin,
        uow: UnitOfWork,
        *args: Any,
        **kwargs: Any,
    ) -> None:
        """Apply the component on the cancel action.

        :param identity: Identity of the user.
        :param request_type: Request type.
        :param action: Action being executed.
        :param topic: Topic of the request.
        :param uow: Unit of work.
        :param args: Additional arguments.
        :param kwargs: Additional keyword arguments.
        """

    def expire(
        self,
        identity: Identity,
        action: OARepoGenericActionMixin,
        uow: UnitOfWork,
        *args: Any,
        **kwargs: Any,
    ) -> None:
        """Apply the component on the expire action.

        :param identity: Identity of the user.
        :param request_type: Request type.
        :param action: Action being executed.
        :param topic: Topic of the request.
        :param uow: Unit of work.
        :param args: Additional arguments.
        :param kwargs: Additional keyword arguments.
        """


class WorkflowTransitionComponent(RequestActionComponent):
    """A component that applies a workflow transition after processing the action.

    When the action is applied, the "status_to" of the request is looked up in
    the workflow transitions for the request and if found, the topic's state is changed
    to the target state.
    """

    def _workflow_transition(self, identity: Identity, action: OARepoGenericActionMixin, uow: UnitOfWork) -> None:
        from oarepo_workflows.proxies import current_oarepo_workflows

        topic = action.topic
        request = action.request
        transitions = current_oarepo_workflows.get_workflow(topic).requests()[request.type.type_id].transitions
        target_state = transitions[action.status_to]
        if target_state and not topic.is_deleted:
            current_oarepo_workflows.set_state(
                identity,
                topic,
                target_state,
                request=request,
                uow=uow,
            )

    @override
    def create(
        self,
        identity: Identity,
        action: OARepoGenericActionMixin,
        uow: UnitOfWork,
        *args: Any,
        **kwargs: Any,
    ) -> None:
        self._workflow_transition(identity, action, uow)

    @override
    def submit(
        self,
        identity: Identity,
        action: OARepoGenericActionMixin,
        uow: UnitOfWork,
        *args: Any,
        **kwargs: Any,
    ) -> None:
        self._workflow_transition(identity, action, uow)

    @override
    def accept(
        self,
        identity: Identity,
        action: OARepoGenericActionMixin,
        uow: UnitOfWork,
        *args: Any,
        **kwargs: Any,
    ) -> None:
        self._workflow_transition(identity, action, uow)

    @override
    def decline(
        self,
        identity: Identity,
        action: OARepoGenericActionMixin,
        uow: UnitOfWork,
        *args: Any,
        **kwargs: Any,
    ) -> None:
        self._workflow_transition(identity, action, uow)

    @override
    def cancel(
        self,
        identity: Identity,
        action: OARepoGenericActionMixin,
        uow: UnitOfWork,
        *args: Any,
        **kwargs: Any,
    ) -> None:
        self._workflow_transition(identity, action, uow)

    @override
    def expire(
        self,
        identity: Identity,
        action: OARepoGenericActionMixin,
        uow: UnitOfWork,
        *args: Any,
        **kwargs: Any,
    ) -> None:
        self._workflow_transition(identity, action, uow)


class AutoAcceptComponent(RequestActionComponent):
    """A component that auto-accepts the request if the receiver has auto-approve enabled.

    IMPORTANT: This component should only be used on the end.
    """

    @override
    def submit(
        self,
        identity: Identity,
        action: OARepoGenericActionMixin,
        uow: UnitOfWork,
        *args: Any,
        **kwargs: Any,
    ) -> None:
        if action.request.status != "submitted":
            return
        receiver_ref = action.request.receiver  # this is <x>proxy, not dict
        if not receiver_ref.reference_dict.get("auto_approve"):
            return

        action_obj = RequestActions.get_action(action.request, "accept")
        if not action_obj.can_execute():
            raise CannotExecuteActionError("accept")
        action_obj.execute(identity, uow, *args, **kwargs)



#
# Copyright (c) 2025 CESNET z.s.p.o.
#
# This file is a part of oarepo-model (see http://github.com/oarepo/oarepo-model).
#
# oarepo-model is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.
#
"""Module providing preset for applying changes to record service config."""

from __future__ import annotations

from typing import TYPE_CHECKING, Any, override

from oarepo_model.customizations import (
    AddToDictionary,
    Customization,
)
from oarepo_model.presets import Preset

from oarepo_requests.services.links import RefEndpointLink

if TYPE_CHECKING:
    from collections.abc import Generator

    from oarepo_model.builder import InvenioModelBuilder
    from oarepo_model.model import InvenioModel


class RequestsServiceConfigPreset(Preset):
    """Preset for record service config class."""

    modifies = ("record_links_item", "record_service_components")

    @override
    def apply(
        self,
        builder: InvenioModelBuilder,
        model: InvenioModel,
        dependencies: dict[str, Any],
    ) -> Generator[Customization]:
        # TODO: AutorequestComponent, RecordSnapshotComponent
        yield AddToDictionary(
            "record_links_item",
            {"requests": RefEndpointLink("requests.search", ref_querystring="topic")},
        )
        yield AddToDictionary(
            "record_links_item",
            {"applicable-requests": RefEndpointLink("requests.applicable_request_types", ref_querystring="topic")},
        )



#
# Copyright (C) 2024 CESNET z.s.p.o.
#
# oarepo-requests is free software; you can redistribute it and/or
# modify it under the terms of the MIT License; see LICENSE file for more
# details.
#
"""Actions for publishing new version draft requests."""

from __future__ import annotations

from typing import TYPE_CHECKING, Any, override

from invenio_i18n import _
from invenio_records_resources.services.uow import RecordCommitOp

from ..utils import get_draft_record_service

if TYPE_CHECKING:
    from flask_principal import Identity
    from invenio_db.uow import UnitOfWork


from .publish_draft import PublishDraftAcceptAction


class PublishNewVersionAcceptAction(PublishDraftAcceptAction):
    """Accept action for publishing new version draft requests."""

    name = _("Publish")

    @override
    def apply(
        self,
        identity: Identity,
        uow: UnitOfWork,
        *args: Any,
        **kwargs: Any,
    ) -> None:
        if "payload" in self.request and "version" in self.request["payload"]:
            topic_service = get_draft_record_service(self.topic)
            self.topic.metadata["version"] = self.request["payload"]["version"]
            uow.register(RecordCommitOp(self.topic, indexer=topic_service.indexer))
        super().apply(identity, uow, *args, **kwargs)



#
# Copyright (C) 2024 CESNET z.s.p.o.
#
# oarepo-requests is free software; you can redistribute it and/or
# modify it under the terms of the MIT License; see LICENSE file for more
# details.
#
"""Mixin for all oarepo actions."""

from __future__ import annotations

from functools import cached_property
from typing import TYPE_CHECKING, Any

from invenio_i18n import _
from invenio_requests.customizations import actions

from oarepo_requests.proxies import current_oarepo_requests
from oarepo_requests.services.permissions.identity import request_active

if TYPE_CHECKING:
    from flask_babel.speaklater import LazyString
    from flask_principal import Identity
    from invenio_db.uow import UnitOfWork
    from invenio_records_resources.records import Record
    from invenio_requests.customizations import RequestAction
    from invenio_requests.records.api import Request

    from oarepo_requests.actions.components import RequestActionComponent
else:
    RequestAction = object


class OARepoGenericActionMixin(RequestAction):
    """Mixin for all oarepo actions."""

    type_id: str
    name: str

    def __init__(self, request: Request):
        """Initialize the action."""
        super().__init__(request)
        self._topic = request.topic.resolve()

    @property
    def topic(self) -> Record:
        """Get the topic."""
        return self._topic

    @topic.setter
    def topic(self, value: Record) -> None:
        """Set the topic."""
        self._topic = value

    @property
    def request(self) -> Request:
        """Get the request."""
        return self._request

    @request.setter
    def request(self, value: Request) -> None:  # type: ignore[reportIncompatibleVariableOverride]
        """Set the topic."""
        self._request = value

    @classmethod
    def stateful_name(cls, identity: Identity, **kwargs: Any) -> str | LazyString:  # noqa ARG003
        """Return the name of the action.

        The name can be a lazy multilingual string and may depend on the state of the action,
        request or identity of the caller.
        """
        return cls.name

    def apply(
        self,
        identity: Identity,
        uow: UnitOfWork,
        *args: Any,
        **kwargs: Any,
    ) -> None:
        """Apply the action to the topic."""

    action_components: tuple[type[RequestActionComponent], ...] = ()

    @cached_property
    def components(self) -> list[RequestActionComponent]:
        """Return a list of components for this action."""
        return [component_cls() for component_cls in current_oarepo_requests.action_components()]

    def execute(self, identity: Identity, uow: UnitOfWork, *args: Any, **kwargs: Any) -> None:
        """Execute the action."""
        was_request_active = request_active in identity.provides
        if not was_request_active:
            identity.provides.add(request_active)
        try:
            self.apply(identity, uow, *args, **kwargs)  # execute the action itself
            super().execute(identity, uow, *args, **kwargs)
            for component in self.components:
                if hasattr(component, self.type_id):
                    getattr(component, self.type_id)(identity, self, uow, *args, **kwargs)
        finally:
            # in case we are not running the actions in isolated state
            if not was_request_active:
                identity.provides.remove(request_active)


class OARepoSubmitAction(OARepoGenericActionMixin, actions.SubmitAction):  # type: ignore[reportIncompatibleVariableOverride]
    """Submit action extended for oarepo requests."""

    type_id = "submit"
    name = _("Submit")


class OARepoDeclineAction(OARepoGenericActionMixin, actions.DeclineAction):  # type: ignore[reportIncompatibleVariableOverride]
    """Decline action extended for oarepo requests."""

    type_id = "decline"
    name = _("Decline")


class OARepoAcceptAction(OARepoGenericActionMixin, actions.AcceptAction):  # type: ignore[reportIncompatibleVariableOverride]
    """Accept action extended for oarepo requests."""

    type_id = "accept"
    name = _("Accept")


class OARepoCancelAction(OARepoGenericActionMixin, actions.CancelAction):  # type: ignore[reportIncompatibleVariableOverride]
    """Cancel action extended for oarepo requests."""

    type_id = "cancel"
    name = _("Cancel")

    status_from = ("created", "submitted")  # type: ignore[reportAssignmentType]
    status_to = "cancelled"



#
# Copyright (C) 2024 CESNET z.s.p.o.
#
# oarepo-requests is free software; you can redistribute it and/or
# modify it under the terms of the MIT License; see LICENSE file for more
# details.
#
"""Publish draft request type."""

from __future__ import annotations

from typing import TYPE_CHECKING, Any, Literal, override

import marshmallow as ma
from invenio_drafts_resources.records import Record as RecordWithDraft
from invenio_i18n import gettext
from invenio_i18n import lazy_gettext as _

from oarepo_requests.actions.publish_draft import (
    PublishDraftAcceptAction,
    PublishDraftDeclineAction,
    PublishDraftSubmitAction,
)

from ..utils import classproperty, get_draft_record_service, search_requests
from .generic import NonDuplicableOARepoRecordRequestType

if TYPE_CHECKING:
    from flask_principal import Identity
    from invenio_records_resources.records import Record
    from invenio_requests.customizations.actions import RequestAction
    from invenio_requests.records.api import Request


from invenio_access.permissions import system_identity
from invenio_requests.records.api import Request

from ..errors import UnresolvedRequestsError


class PublishRequestType(NonDuplicableOARepoRecordRequestType):
    """Publish draft request type."""

    description = _("Request to publish a draft")  # type: ignore[reportAssignmentType]
    allowed_on_published = False
    receiver_can_be_none = True
    editable = False

    @classproperty
    def available_actions(cls) -> dict[str, type[RequestAction]]:  # noqa N805
        """Return available actions for the request type."""
        return {
            **super().available_actions,
            "submit": PublishDraftSubmitAction,
            "accept": PublishDraftAcceptAction,
            "decline": PublishDraftDeclineAction,
        }

    @override
    def can_create(
        self,
        identity: Identity,
        data: dict,
        receiver: dict[str, str],
        topic: Record,
        creator: dict[str, str],
        *args: Any,
        **kwargs: Any,
    ) -> None:
        """Check if the request can be created."""
        if not isinstance(topic, RecordWithDraft):
            raise TypeError(f"Topic type {type(topic)} does not support drafts")
        self.assert_no_pending_requests(topic)
        super().can_create(identity, data, receiver, topic, creator, *args, **kwargs)
        self.validate_topic(identity, topic)

    @classmethod
    @override
    def is_applicable_to(cls, identity: Identity, topic: Record, *args: Any, **kwargs: Any) -> bool:
        """Check if the request type is applicable to the topic."""
        if not isinstance(topic, RecordWithDraft):
            raise TypeError(f"Topic type {type(topic)} does not support drafts")
        return super().is_applicable_to(identity, topic, *args, **kwargs)

    def assert_no_pending_requests(
        self,
        topic: Record,
    ) -> None:
        """Assert that there are no pending requests on the topic."""
        requests = search_requests(system_identity, topic)

        for result in requests._results:  # noqa SLF001
            # note: we can not use solely the result.is_open because changes may not be committed yet
            # to opensearch index. That's why we need to get the record from DB and re-check.
            if (
                result.is_open
                and result.type != self.type_id
                and Request.get_record(result.uuid)["status"]
                in (
                    "submitted",
                    "created",
                )
            ):
                raise UnresolvedRequestsError(action=str(self.name))

    @classmethod
    def validate_topic(cls, identity: Identity, topic: Record) -> None:
        """Validate the topic.

        :param: identity: identity of the caller
        :param: topic: topic of the request

        :raises: ValidationError: if the topic is not valid
        """
        topic_service = get_draft_record_service(topic)
        topic_service.validate_draft(identity, topic["id"])

        # if files support is enabled for this topic, check if there are any files
        if hasattr(topic, "files"):
            can_toggle_files = topic_service.check_permission(identity, "manage_files", record=topic)
            draft_files = topic.files  # type: ignore[reportAttributeAccessIssue]
            if draft_files.enabled and not draft_files.items():
                if can_toggle_files:
                    my_message = gettext(
                        "Missing uploaded files. To disable files for this record please mark it as metadata-only."
                    )
                else:
                    my_message = gettext("Missing uploaded files.")

                raise ma.ValidationError({"files.enabled": [my_message]})

    @classmethod
    def topic_type(cls, topic: Record) -> Literal["initial", "new_version", "metadata", "published"]:
        """Return publish status type of the topic."""
        if not isinstance(topic, RecordWithDraft):
            raise TypeError(f"Topic type {type(topic)} does not support drafts")
        index = topic.versions.index
        is_latest = topic.versions.is_latest
        is_draft = topic.is_draft

        if not is_draft:
            return "published"

        if index == 1 and not is_latest:
            return "initial"
        if index > 1 and not is_latest:
            return "new_version"
        return "metadata"



#
# Copyright (c) 2025 CESNET z.s.p.o.
#
# This file is a part of oarepo-requests (see https://github.com/oarepo/oarepo-requests).
#
# oarepo-requests is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.
#
"""Links Module for requests."""

from __future__ import annotations

from typing import TYPE_CHECKING, Any, override

from invenio_records_resources.services.base.links import EndpointLink
from invenio_requests.resolvers.registry import ResolverRegistry

if TYPE_CHECKING:
    from collections.abc import Callable


class RefEndpointLink(EndpointLink):
    """Endpoint link that adds reference query string parameter to the URL."""

    @override
    def __init__(
        self,
        endpoint: str,
        *,
        when: Callable | None = None,
        vars: Callable | None = None,
        params: list[str] | None = None,
        ref_querystring: str,
    ):
        """Construct.

        :param endpoint: str. endpoint of the URL
        :param when: fn(obj, dict) -> bool, when the URL should be rendered
        :param vars: fn(obj, dict), mutate dict in preparation for expansion
        :param params: list, parameters (excluding querystrings) used for expansion
        :param ref_querystring: str, name of the querystring parameter to be added to the URL
        """
        super().__init__(endpoint, when, vars, params)
        self._ref_querystring = ref_querystring

    @override
    def expand(self, obj: Any, context: dict[str, Any]) -> str:
        """Expand the endpoint."""
        ret = super().expand(obj, context)
        topic_ref = ResolverRegistry.reference_entity(obj)
        if topic_ref is None:
            raise ValueError(f"Can't create link with unresolvable entity {obj}")
        return f"{ret}?{self._ref_querystring}={next(iter(topic_ref.keys()))}:{next(iter(topic_ref.values()))}"



#
# Copyright (C) 2024 CESNET z.s.p.o.
#
# oarepo-requests is free software; you can redistribute it and/or
# modify it under the terms of the MIT License; see LICENSE file for more
# details.
#
"""Blueprints for the app and events views."""

from __future__ import annotations

from pathlib import Path
from typing import TYPE_CHECKING

from flask import Blueprint

if TYPE_CHECKING:
    from flask import Flask


def create_notifications(app: Flask) -> Blueprint:  # noqa ARG001
    """Register blueprint routes on app."""
    return Blueprint(
        "oarepo_notifications",
        __name__,
        template_folder=Path(__file__).parent.parent / "templates",
    )



#
# Copyright (C) 2024 CESNET z.s.p.o.
#
# oarepo-requests is free software; you can redistribute it and/or
# modify it under the terms of the MIT License; see LICENSE file for more
# details.
#
"""Base request type for OARepo requests."""

from __future__ import annotations

from typing import TYPE_CHECKING, Any, cast

from invenio_records_resources.services.errors import PermissionDeniedError
from invenio_requests.customizations import RequestType
from invenio_requests.customizations.states import RequestState
from invenio_requests.proxies import current_requests_service

from oarepo_requests.errors import OpenRequestAlreadyExistsError
from oarepo_requests.notifications.builders.comment import (
    CommentRequestEventCreateNotificationBuilder,
)
from oarepo_requests.utils import classproperty, open_request_exists

from ..actions.generic import (
    OARepoAcceptAction,
    OARepoCancelAction,
    OARepoDeclineAction,
    OARepoSubmitAction,
)
from ..utils import (
    has_rights_to_accept_request,
    has_rights_to_submit_request,
    is_auto_approved,
    request_identity_matches,
)
from .ref_types import ModelRefTypes, ReceiverRefTypes

if TYPE_CHECKING:
    from collections.abc import Callable

    from flask_babel.speaklater import LazyString
    from flask_principal import Identity
    from invenio_records_resources.records import Record
    from invenio_requests.customizations.actions import RequestAction
    from invenio_requests.records.api import Request
    from marshmallow.schema import Schema


class OARepoRequestType(RequestType):
    """Base request type for OARepo requests."""

    description: str = ""

    dangerous = False

    allowed_on_draft = True
    allowed_on_published = True

    editable: bool | None = None
    """Whether the request type can be edited multiple times before it is submitted."""

    allowed_receiver_ref_types = ReceiverRefTypes()  # type: ignore[reportAssignmentType]

    comment_notification_builder = CommentRequestEventCreateNotificationBuilder

    @classproperty
    def has_form(cls) -> bool:  # noqa N805
        """Return whether the request type has a form."""
        return hasattr(cls, "form")

    @classproperty
    def is_editable(cls) -> bool:  # noqa N805
        """Return whether the request type is editable."""
        if cls.editable is not None:
            return cls.editable
        return cls.has_form

    @classproperty
    def available_statuses(cls) -> dict[str, RequestState]:  # type: ignore[reportIncompatibleVariableOverride] # noqa N805
        """Return available statuses for the request type.

        The status (open, closed, undefined) are used for request filtering.
        """
        return {**super().available_statuses, "created": RequestState.OPEN}

    @classmethod
    def _create_marshmallow_schema(cls) -> type[Schema]:
        """Create a marshmallow schema for this request type with required payload field."""
        schema = super()._create_marshmallow_schema()
        # TODO: idk why .fields
        if cls.payload_schema is not None and hasattr(schema, "fields") and "payload" in schema.fields:  # type: ignore[reportAttributeAccessIssue]
            schema.fields["payload"].required = True  # type: ignore[reportAttributeAccessIssue]

        return cast("type[Schema]", schema)

    def can_create(
        self,
        identity: Identity,
        data: dict[str, Any],  # noqa ARG002
        receiver: dict[str, str],  # noqa ARG002
        topic: Record,
        creator: dict[str, str],  # noqa ARG002
        *args: Any,  # noqa ARG002
        **kwargs: Any,
    ) -> None:
        """Check if the request can be created.

        :param identity:        identity of the caller
        :param data:            data of the request
        :param receiver:        receiver of the request
        :param topic:           topic of the request
        :param creator:         creator of the request
        :param args:            additional arguments
        :param kwargs:          additional keyword arguments
        """
        current_requests_service.require_permission(identity, "create", record=topic, request_type=self, **kwargs)

    @classmethod
    def is_applicable_to(cls, identity: Identity, topic: Record, *args: Any, **kwargs: Any) -> bool:  # noqa ARG002
        """Check if the request type is applicable to the topic.

        Used for checking whether there is any situation where the client can create
        a request of this type it's different to just using can create with no receiver
        and data because that checks specifically for situation without them while this
        method is used to check whether there is a possible situation a user might create
        this request eg. for the purpose of serializing a link on associated record
        """
        return cast(
            "bool",
            current_requests_service.check_permission(identity, "create", record=topic, request_type=cls, **kwargs),
        )

    @classproperty
    def available_actions(  # type: ignore[reportIncompatibleVariableOverride]
        self,
    ) -> dict[str, type[RequestAction]]:
        """Return available actions for the request type."""
        return {
            **super().available_actions,
            "submit": OARepoSubmitAction,
            "accept": OARepoAcceptAction,
            "decline": OARepoDeclineAction,
            "cancel": OARepoCancelAction,
        }

    # TODO: move these to RecordRequestType too?
    def stateful_name(
        self,
        identity: Identity,  # noqa ARG002
        *,
        topic: Record,  # noqa ARG002
        request: Request | None = None,  # noqa ARG002
        **kwargs: Any,  # noqa ARG002
    ) -> str | LazyString:
        """Return the name of the request that reflects its current state.

        :param identity:        identity of the caller
        :param request:         the request
        :param topic:           resolved request's topic
        """
        return self.name

    def stateful_description(
        self,
        identity: Identity,  # noqa ARG002
        *,
        topic: Record,  # noqa ARG002
        request: Request | None = None,  # noqa ARG002
        **kwargs: Any,  # noqa ARG002
    ) -> str | LazyString:
        """Return the description of the request that reflects its current state.

        :param identity:        identity of the caller
        :param request:         the request
        :param topic:           resolved request's topic
        """
        return self.description

    def string_by_state(  # noqa C901, PLR0913, PLR0911
        self,
        identity: Identity,
        *,
        topic: Record,
        request: Request | None = None,
        # strings
        create: (str | LazyString | Callable[[Identity, Record, Request | None], str | LazyString]),
        create_autoapproved: (str | LazyString | Callable[[Identity, Record, Request | None], str | LazyString]),
        submit: (str | LazyString | Callable[[Identity, Record, Request | None], str | LazyString]),
        submitted_receiver: (str | LazyString | Callable[[Identity, Record, Request | None], str | LazyString]),
        submitted_creator: (str | LazyString | Callable[[Identity, Record, Request | None], str | LazyString]),
        submitted_others: (str | LazyString | Callable[[Identity, Record, Request | None], str | LazyString]),
        accepted: (str | LazyString | Callable[[Identity, Record, Request | None], str | LazyString]),
        declined: (str | LazyString | Callable[[Identity, Record, Request | None], str | LazyString]),
        cancelled: (str | LazyString | Callable[[Identity, Record, Request | None], str | LazyString]),
        created: (str | LazyString | Callable[[Identity, Record, Request | None], str | LazyString]),
    ) -> str | LazyString:
        """Return a string that varies by the state of the request.

        :param create:        string to be used on request type if user can create a request
        :param create_autoapproved: string to be used on request type if user can create a request
                                    and the request is auto approved
        :param submit:        string to be used on request type if user can submit a request
        :param accept_decline: string to be used on request type if user can accept or decline a request
        :param view:          string to be used on request type if user can view a request
        """

        def get_string(
            string: str | LazyString | Callable[[Identity, Record, Request | None], str | LazyString],
            identity: Identity,
            topic: Record,
            request: Request | None = None,
        ) -> str | LazyString:
            if callable(string):
                return string(identity, topic, request)
            return string

        if request:
            match request.status:
                case "submitted":
                    if has_rights_to_accept_request(request, identity):
                        return get_string(submitted_receiver, identity, topic, request)
                    if request_identity_matches(request.created_by, identity):
                        return get_string(submitted_creator, identity, topic, request)
                    return get_string(submitted_others, identity, topic, request)
                case "accepted":
                    return get_string(accepted, identity, topic, request)
                case "declined":
                    return get_string(declined, identity, topic, request)
                case "cancelled":
                    return get_string(cancelled, identity, topic, request)
                case "created":
                    if has_rights_to_submit_request(request, identity):
                        return get_string(submit, identity, topic, request)
                    return get_string(created, identity, topic, request)
                case _:
                    return f'Unknown label for status "{request.status}" in "{__file__}"'

        if is_auto_approved(self, identity=identity, topic=topic):
            return get_string(create_autoapproved, identity, topic, request)
        return get_string(create, identity, topic, request)


class OARepoRecordRequestType(OARepoRequestType):
    """Base request type for OARepo requests that can be created on a record."""

    allowed_topic_ref_types = ModelRefTypes()  # type: ignore[reportAssignmentType]

    @classmethod
    def _allowed_by_publication_status(cls, record: Record) -> bool:
        if not hasattr(record, "publication_status"):
            return bool(cls.allowed_on_published)
        return bool(
            (cls.allowed_on_draft and record.publication_status == "draft")  # type: ignore[reportAttributeAccessIssue]
            or (cls.allowed_on_published and record.publication_status == "published")  # type: ignore[reportAttributeAccessIssue]
        )

    def can_create(
        self,
        identity: Identity,
        data: dict[str, Any],
        receiver: dict[str, str],
        topic: Record,
        creator: dict[str, str],
        *args: Any,
        **kwargs: Any,
    ) -> None:
        """Check if the request can be created.

        :param identity:        identity of the caller
        :param data:            data of the request
        :param receiver:        receiver of the request
        :param topic:           topic of the request
        :param creator:         creator of the request
        :param args:            additional arguments
        :param kwargs:          additional keyword arguments
        """
        if not self._allowed_by_publication_status(record=topic):
            raise PermissionDeniedError("create")
        super().can_create(identity, data, receiver, topic, creator, *args, **kwargs)

    @classmethod
    def is_applicable_to(cls, identity: Identity, topic: Record, *args: Any, **kwargs: Any) -> bool:
        """Check if the request type is applicable to the topic.

        Used for checking whether there is any situation where the client can create
        a request of this type it's different to just using can create with no receiver
        and data because that checks specifically for situation without them while this
        method is used to check whether there is a possible situation a user might create
        this request eg. for the purpose of serializing a link on associated record
        """
        if not cls._allowed_by_publication_status(record=topic):
            return False
        return super().is_applicable_to(identity, topic, *args, **kwargs)


class NonDuplicableOARepoRecordRequestType(OARepoRecordRequestType):
    """Base request type for OARepo requests that cannot be duplicated.

    This means that on a single topic there can be only one open request of this type.
    """

    def can_create(
        self,
        identity: Identity,
        data: dict,
        receiver: dict[str, str],
        topic: Record,
        creator: dict[str, str],
        *args: Any,
        **kwargs: Any,
    ) -> None:
        """Check if the request can be created.

        :param identity:        identity of the caller
        :param data:            data of the request
        :param receiver:        receiver of the request
        :param topic:           topic of the request
        :param creator:         creator of the request
        :param args:            additional arguments
        :param kwargs:          additional keyword arguments
        """
        if open_request_exists(topic, self.type_id):
            raise OpenRequestAlreadyExistsError(self, topic)
        super().can_create(identity, data, receiver, topic, creator, *args, **kwargs)

    @classmethod
    def is_applicable_to(cls, identity: Identity, topic: Record, *args: Any, **kwargs: Any) -> bool:
        """Check if the request type is applicable to the topic."""
        if open_request_exists(topic, cls.type_id):
            return False
        return super().is_applicable_to(identity, topic, *args, **kwargs)



#
# Copyright (C) 2024 CESNET z.s.p.o.
#
# oarepo-requests is free software; you can redistribute it and/or
# modify it under the terms of the MIT License; see LICENSE file for more
# details.
#
"""Request for deleting published record."""

from __future__ import annotations

from typing import TYPE_CHECKING, Any, ClassVar, override

import marshmallow as ma
from invenio_i18n import gettext
from invenio_i18n import lazy_gettext as _

from oarepo_requests.actions.delete_published_record import (
    DeletePublishedRecordAcceptAction,
    DeletePublishedRecordDeclineAction,
    DeletePublishedRecordSubmitAction,
)

from ..utils import (
    classproperty,
    is_auto_approved,
    open_request_exists,
    request_identity_matches,
)
from .generic import NonDuplicableOARepoRecordRequestType

if TYPE_CHECKING:
    from collections.abc import Mapping

    from flask_babel.speaklater import LazyString
    from flask_principal import Identity
    from invenio_records_resources.records import Record
    from invenio_requests.customizations.actions import RequestAction
    from invenio_requests.records.api import Request

    from ..utils import JsonValue


class DeletePublishedRecordRequestType(NonDuplicableOARepoRecordRequestType):
    """Request type for requesting deletion of a published record."""

    type_id = "delete_published_record"
    name = _("Delete record")  # type: ignore[reportAssignmentType]
    description = _("Request deletion of published record")  # type: ignore[reportAssignmentType]
    allowed_on_draft = False
    editable = False
    dangerous = True
    receiver_can_be_none = True

    payload_schema: Mapping[str, ma.fields.Field] | None = {
        "removal_reason": ma.fields.Str(required=True),
        "note": ma.fields.Str(),
    }

    @classmethod
    def is_applicable_to(cls, identity: Identity, topic: Record, *args: Any, **kwargs: Any) -> bool:
        """Check if the request type is applicable to the topic."""
        if open_request_exists(topic, cls.type_id):
            return False
        return super().is_applicable_to(identity, topic, *args, **kwargs)

    @classproperty
    def available_actions(cls) -> dict[str, type[RequestAction]]:  # noqa N805
        """Return available actions for the request type."""
        return {
            **super().available_actions,
            "submit": DeletePublishedRecordSubmitAction,
            "accept": DeletePublishedRecordAcceptAction,
            "decline": DeletePublishedRecordDeclineAction,
        }

    # TODO: used in ui
    form: ClassVar[JsonValue] = [
        {
            "section": "",
            "fields": [
                {
                    "field": "removal_reason",
                    "ui_widget": "Input",
                    "props": {
                        "label": _("Removal Reason"),
                        "placeholder": _("Write down the removal reason."),
                        "required": True,
                    },
                },
                {
                    "section": "",
                    "field": "note",
                    "ui_widget": "Input",
                    "props": {
                        "label": _("Note"),
                        "placeholder": _("Write down the additional note."),
                        "required": False,
                    },
                },
            ],
        }
    ]

    @override
    def stateful_name(
        self,
        identity: Identity,
        *,
        topic: Record,
        request: Request | None = None,
        **kwargs: Any,
    ) -> str | LazyString:
        """Return the stateful name of the request."""
        if is_auto_approved(self, identity=identity, topic=topic):
            return self.name
        if not request:
            return gettext("Request record deletion")
        match request.status:
            case "submitted":
                return gettext("Record deletion requested")
            case _:
                return gettext("Request record deletion")

    @override
    def stateful_description(  # noqa PLR0911
        self,
        identity: Identity,
        *,
        topic: Record,
        request: Request | None = None,
        **kwargs: Any,
    ) -> str | LazyString:
        """Return the stateful description of the request."""
        if is_auto_approved(self, identity=identity, topic=topic):
            return gettext("Click to permanently delete the record.")

        if not request:
            return gettext("Request permission to delete the record.")
        match request.status:
            case "submitted":
                if request_identity_matches(request.created_by, identity):
                    return gettext(
                        "Permission to delete record requested. You will be notified about the decision by email."
                    )
                if request_identity_matches(request.receiver, identity):
                    return gettext(
                        "You have been asked to approve the request to permanently delete the record. "
                        "You can approve or reject the request."
                    )
                return gettext("Permission to delete record (including files) requested. ")
            case _:
                if request_identity_matches(request.created_by, identity):
                    return gettext("Submit request to get permission to delete the record.")
                return gettext("You do not have permission to delete the record.")



#
# Copyright (C) 2025 CESNET z.s.p.o.
#
# oarepo-requests is free software; you can redistribute it and/or
# modify it under the terms of the MIT License; see LICENSE file for more
# details.
#
"""OARepo requests facets param interpreters."""

from __future__ import annotations

from functools import partial
from typing import TYPE_CHECKING, override

from invenio_records_resources.services.records.params import (
    FilterParam,
    ParamInterpreter,
)
from invenio_requests.services.requests.params import IsOpenParam
from invenio_search.engine import dsl

if TYPE_CHECKING:
    from flask_principal import Identity
    from invenio_records_resources.services import SearchOptions
    from invenio_search.api import RecordsSearchV2


class RequestOwnerFilterParam(FilterParam):
    """Filter requests by owner."""

    # params have to be dict; support for pop operation required
    @override
    def apply(self, identity: Identity, search: RecordsSearchV2, params: dict[str, str]) -> RecordsSearchV2:  # type: ignore[reportIncompatibleMethodOverride]
        """Apply the filter to the search."""
        value = params.pop(self.param_name, None)
        if value is not None:
            search = search.filter("term", **{self.field_name: identity.id})
        return search


class RequestAllAvailableFilterParam(ParamInterpreter):
    """A no-op filter that returns all requests that are readable by the current user."""

    def __init__(self, param_name: str, config: type[SearchOptions]) -> None:
        """Initialize the filter."""
        self.param_name = param_name
        super().__init__(config)

    @classmethod
    def factory(cls, param: str) -> partial[ParamInterpreter]:
        """Create a new filter parameter."""
        return partial(cls, param)

    @override
    def apply(self, identity: Identity, search: RecordsSearchV2, params: dict[str, str]) -> RecordsSearchV2:  # type: ignore[reportIncompatibleMethodOverride]
        """Apply the filter to the search - does nothing."""
        params.pop(self.param_name, None)
        return search


class RequestNotOwnerFilterParam(FilterParam):
    """Filter requests that are not owned by the current user.

    Note: invenio still does check that the user has the right to see the request,
    so this is just a filter to narrow down the search to requests, that the user
    can approve.
    """

    @override
    def apply(self, identity: Identity, search: RecordsSearchV2, params: dict[str, str]) -> RecordsSearchV2:  # type: ignore[reportIncompatibleMethodOverride]
        """Apply the filter to the search."""
        value = params.pop(self.param_name, None)
        if value is not None:
            search = search.filter(dsl.query.Bool(must_not=[dsl.Q("term", **{self.field_name: identity.id})]))
        return search


class IsClosedParam(IsOpenParam):
    """Get just the closed requests."""

    @override
    def apply(self, identity: Identity, search: RecordsSearchV2, params: dict[str, str]) -> RecordsSearchV2:  # type: ignore[reportIncompatibleMethodOverride]
        """Evaluate the is_closed parameter on the search."""
        if params.get("is_closed") is True:
            search = search.filter("term", **{self.field_name: True})
        elif params.get("is_closed") is False:
            search = search.filter("term", **{self.field_name: False})
        return search



#
# Copyright (C) 2024 CESNET z.s.p.o.
#
# oarepo-requests is free software; you can redistribute it and/or
# modify it under the terms of the MIT License; see LICENSE file for more
# details.
#
"""Generator is triggered when workflow action is being performed."""

from __future__ import annotations

from typing import TYPE_CHECKING, Any, override

from invenio_records_permissions.generators import Generator
from invenio_search.engine import dsl

from oarepo_requests.services.permissions.identity import request_active

if TYPE_CHECKING:
    from flask_principal import Need


class RequestActive(Generator):
    """A generator that requires that a request is being handled.

    This is useful for example when a caller identity should have greater permissions
    when calling an action from within a request.
    """

    @override
    def needs(self, **context: Any) -> list[Need]:
        """Return the needs required for the action."""
        return [request_active]

    @override
    def query_filter(self, **context: Any) -> dsl.query.Query:
        """Return the query filter for the action."""
        return dsl.Q("match_none")



#
# Copyright (C) 2024 CESNET z.s.p.o.
#
# oarepo-requests is free software; you can redistribute it and/or
# modify it under the terms of the MIT License; see LICENSE file for more
# details.
#
"""Configuration for the oarepo request service."""

from __future__ import annotations

import logging

from invenio_requests.services import RequestsServiceConfig

from oarepo_requests.services.search import EnhancedRequestSearchOptions

log = logging.getLogger(__name__)


class OARepoRequestsServiceConfig(RequestsServiceConfig):
    """Configuration for the oarepo request service."""

    search = EnhancedRequestSearchOptions



#
# Copyright (C) 2024 CESNET z.s.p.o.
#
# oarepo-requests is free software; you can redistribute it and/or
# modify it under the terms of the MIT License; see LICENSE file for more
# details.
#
"""Default workflow receiver function."""

from __future__ import annotations

from typing import TYPE_CHECKING, Any, cast

from oarepo_workflows.errors import RequestTypeNotInWorkflowError

from oarepo_requests.errors import ReceiverNonReferencableError

if TYPE_CHECKING:
    from collections.abc import Mapping

    from invenio_records_resources.records.api import Record
    from invenio_requests.customizations.request_types import RequestType
    from oarepo_workflows import WorkflowRequest


def default_workflow_receiver_function(
    record: Record,
    request_type: RequestType,
    **kwargs: Any,  # i suppose we can't have requests with None topic
) -> Mapping[str, str] | None:
    """Get the receiver of the request.

    This function is called by oarepo-requests when a new request is created. It should
    return the receiver of the request. The receiver is the entity that is responsible for
    accepting/declining the request.
    """
    from oarepo_workflows.proxies import current_oarepo_workflows

    workflow = current_oarepo_workflows.get_workflow(record)
    if not workflow:
        return None  # exception?

    try:
        request: WorkflowRequest = workflow.requests().requests_by_id[request_type.type_id]
    except KeyError as e:
        raise RequestTypeNotInWorkflowError(
            request_type.type_id, current_oarepo_workflows.get_workflow(record).code
        ) from e

    receiver = request.recipient_entity_reference(record=record, request_type=request_type, **kwargs)
    if not request_type.receiver_can_be_none and not receiver:
        raise ReceiverNonReferencableError(request_type=request_type, record=record, **kwargs)
    return cast("Mapping[str, str]", receiver)  # TODO: unknown linter issue, possibly a bug



#
# Copyright (C) 2024 CESNET z.s.p.o.
#
# oarepo-requests is free software; you can redistribute it and/or
# modify it under the terms of the MIT License; see LICENSE file for more
# details.
#
"""Errors raised by oarepo-requests."""

from __future__ import annotations

import json
from typing import TYPE_CHECKING, Any, cast, override

from flask import g
from flask_resources import (
    HTTPJSONException,
)
from flask_resources.serializers.json import JSONEncoder
from invenio_i18n import gettext
from invenio_i18n import lazy_gettext as _
from invenio_requests.errors import CannotExecuteActionError

if TYPE_CHECKING:
    from invenio_records_resources.records import Record
    from invenio_requests.customizations import RequestType


class CustomHTTPJSONException(HTTPJSONException):
    """Custom HTTP Exception delivering JSON error responses with an error_type."""

    def __init__(
        self,
        code: int | None = None,
        errors: list[dict[str, Any]] | None = None,
        topic_errors: list[dict[str, Any]] | None = None,
        request_payload_errors: list[dict[str, Any]] | None = None,
        **kwargs: Any,
    ) -> None:
        """Initialize CustomHTTPJSONException."""
        super().__init__(code=code, errors=errors, **kwargs)
        self.topic_errors = topic_errors or []
        self.request_payload_errors = request_payload_errors or []
        self.extra_kwargs = kwargs  # Save all kwargs

    @override
    def get_body(self, environ: Any | None = None, scope: Any | None = None) -> str:
        """Get the request body."""
        body = {"status": self.code, "message": self.get_description(environ)}

        errors = self.get_errors()
        if errors:
            body["errors"] = errors

        if self.topic_errors:
            body["topic_errors"] = self.topic_errors

        if self.request_payload_errors:
            body["request_payload_errors"] = self.request_payload_errors

        if self.code and (self.code >= 500) and hasattr(g, "sentry_event_id"):  # noqa PLR2004
            body["error_id"] = str(g.sentry_event_id)

        return json.dumps(body, cls=JSONEncoder)


class OpenRequestAlreadyExistsError(CannotExecuteActionError):
    """An open request already exists."""

    def __init__(self, request_type: RequestType, record: Record) -> None:
        """Initialize the exception."""
        self.request_type = request_type
        self.record = record

    def __str__(self):
        """Return str representation."""
        return self.description

    @property
    def description(self) -> str:
        """Exception's description."""
        return cast(
            "str",
            gettext("There is already an open request of %(request_type)s on %(record_id)s.")
            % {
                "request_type": self.request_type.name,
                "record_id": self.record.id,
            },
        )


class UnresolvedRequestsError(CannotExecuteActionError):
    """There were unresolved requests before an action could proceed."""

    def __init__(self, action: str, reason: str | None = None) -> None:
        """Initialize the exception."""
        self.action = action
        self.reason = reason or _("All open requests must be closed first.")

    def __str__(self):
        """Return str(self)."""
        return gettext("Cannot %(action)s: %(reason)s") % {
            "action": self.action.lower(),
            "reason": self.reason,
        }


class UnknownRequestTypeError(Exception):
    """Exception raised when user tries to create a request with an unknown request type."""

    def __init__(self, request_type: str) -> None:
        """Initialize the exception."""
        self.request_type = request_type

    @property
    def description(self) -> str:
        """Exception's description."""
        return gettext("Unknown request type %(request_type)s.") % {  # type: ignore[no-any-return]
            "request_type": self.request_type,
        }


class ReceiverNonReferencableError(Exception):
    """Raised when receiver is required but could not be estimated from the record/caller."""

    def __init__(self, request_type: RequestType, record: Record, **kwargs: Any) -> None:
        """Initialize the exception."""
        self.request_type = request_type
        self.record = record
        self.kwargs = kwargs

    @property
    def description(self) -> str:
        """Exception's description."""
        message = gettext(
            "Receiver for request type %(request_type)s is required but wasn't successfully "
            "referenced on record %(record_id)s."
        ) % {
            "request_type": self.request_type,
            "record_id": self.record["id"],
        }
        if self.kwargs:
            message += gettext("\n Additional keyword arguments:")
            message += f"\n{', '.join(self.kwargs)}"
        return message  # type: ignore[no-any-return]


class VersionAlreadyExists(CustomHTTPJSONException):
    """Exception raised when a version tag already exists."""

    def __init__(self) -> None:
        """Initialize the exception."""
        description = gettext("There is already a record version with this version tag.")
        request_payload_errors = [
            {
                "field": "payload.version",
                "messages": [
                    gettext(
                        "There is already a record version with this version tag. Please use a different version tag."
                    )
                ],
            }
        ]
        super().__init__(
            code=400,
            description=description,
            request_payload_errors=request_payload_errors,
        )



#
# Copyright (C) 2024 CESNET z.s.p.o.
#
# oarepo-requests is free software; you can redistribute it and/or
# modify it under the terms of the MIT License; see LICENSE file for more
# details.
#
"""EditPublishedRecordRequestType."""

from __future__ import annotations

from typing import TYPE_CHECKING, Any, override

from invenio_drafts_resources.records import Record as RecordWithDraft
from invenio_i18n import gettext
from invenio_i18n import lazy_gettext as _
from invenio_requests.records.api import Request
from oarepo_runtime.records.drafts import has_draft

from oarepo_requests.actions.edit_topic import EditTopicAcceptAction

from ..utils import classproperty, is_auto_approved, request_identity_matches
from .generic import NonDuplicableOARepoRecordRequestType

if TYPE_CHECKING:
    from flask_babel.speaklater import LazyString
    from flask_principal import Identity
    from invenio_records_resources.records import Record
    from invenio_requests.customizations.actions import RequestAction
    from invenio_requests.records.api import Request


class EditPublishedRecordRequestType(NonDuplicableOARepoRecordRequestType):
    """Request type for requesting edit of a published record.

    This request type is used to request edit access to a published record. This access
    is restricted to the metadata of the record, not to the files.
    """

    type_id = "edit_published_record"
    name = _("Edit metadata")
    description = _("Request re-opening of published record")  # type: ignore[reportAssignmentType]
    allowed_on_draft = False
    receiver_can_be_none = True

    @classproperty
    def available_actions(cls) -> dict[str, type[RequestAction]]:  # noqa N805
        """Return available actions for the request type."""
        return {
            **super().available_actions,
            "accept": EditTopicAcceptAction,
        }

    def can_create(
        self,
        identity: Identity,
        data: dict[str, Any],
        receiver: dict[str, str],
        topic: Record,
        creator: dict[str, str],
        *args: Any,
        **kwargs: Any,
    ) -> None:
        """Check if the request can be created.

        :param identity:        identity of the caller
        :param data:            data of the request
        :param receiver:        receiver of the request
        :param topic:           topic of the request
        :param creator:         creator of the request
        :param args:            additional arguments
        :param kwargs:          additional keyword arguments
        """
        if not isinstance(topic, RecordWithDraft):
            raise TypeError(gettext("Trying to create edit request on record without draft support"))
        if has_draft(topic):
            raise ValueError(gettext("Trying to create edit request on record with draft"))
        super().can_create(identity, data, receiver, topic, creator, *args, **kwargs)

    @classmethod
    def is_applicable_to(cls, identity: Identity, topic: Record, *args: Any, **kwargs: Any) -> bool:
        """Check if the request type is applicable to the topic."""
        if not isinstance(topic, RecordWithDraft):
            return False
        # if already editing metadata or a new version, we don't want to create a new request
        if has_draft(topic):
            return False
        return super().is_applicable_to(identity, topic, *args, **kwargs)

    # TODO: used in ui
    @override
    def stateful_name(
        self,
        identity: Identity,
        *,
        topic: Record,
        request: Request | None = None,
        **kwargs: Any,
    ) -> str | LazyString:
        """Return the stateful name of the request."""
        if is_auto_approved(self, identity=identity, topic=topic):
            return self.name
        if not request:
            return gettext("Request edit access")
        match request.status:
            case "submitted":
                return gettext("Edit access requested")
            case _:
                return gettext("Request edit access")

    @override
    def stateful_description(
        self,
        identity: Identity,
        *,
        topic: Record,
        request: Request | None = None,
        **kwargs: Any,
    ) -> str | LazyString:
        """Return the stateful description of the request."""
        if is_auto_approved(self, identity=identity, topic=topic):
            return gettext("Click to start editing the metadata of the record.")

        if not request:
            return gettext("Request edit access to the record. You will be notified about the decision by email.")
        match request.status:
            case "submitted":
                if request_identity_matches(request.created_by, identity):
                    return gettext("Edit access requested. You will be notified about the decision by email.")
                if request_identity_matches(request.receiver, identity):
                    return gettext("You have been requested to grant edit access to the record.")
                return gettext("Edit access requested.")
            case _:
                return gettext("Request edit access to the record. You will be notified about the decision by email.")



#
# Copyright (C) 2025 CESNET z.s.p.o.
#
# oarepo-requests is free software; you can redistribute it and/or
# modify it under the terms of the MIT License; see LICENSE file for more
# details.
#
"""Notification recipients generators in oarepo-requests."""

from __future__ import annotations

from typing import TYPE_CHECKING, Any

from invenio_access.permissions import system_identity
from invenio_notifications.models import Notification, Recipient
from invenio_notifications.services.generators import RecipientGenerator
from invenio_records.dictutils import dict_lookup
from invenio_requests import current_events_service
from invenio_search.engine import dsl
from invenio_users_resources.proxies import current_users_service

from oarepo_requests.notifications.generators.context import (
    NotificationCtxWithReference,
)
from oarepo_requests.proxies import current_notification_recipient_generators_registry

if TYPE_CHECKING:
    from invenio_accounts.models import User


def _extract_user_email_data(user: User) -> dict[str, Any]:
    return {
        "id": user.id,
        "preferences": dict(user.preferences) if user.preferences else user.preferences,
        "email": user.email,
    }


class EntityRecipientGenerator(RecipientGenerator):
    """Recipient generator working as handler for generic entity."""

    def __init__(self, key: str):
        """Initialize the generator."""
        self.key = key

    def __call__(self, notification: Notification, recipients: dict[str, Recipient]):
        """Generate recipients for the given entity."""
        if not isinstance(notification.context, NotificationCtxWithReference):
            raise TypeError(
                "Using EntityRecipientGenerator on notification without saved reference dictionary to the entity."
            )
        entity_type = notification.context.get_reference_type(self.key)
        generator = current_notification_recipient_generators_registry[entity_type](self.key, notification)
        generator(notification, recipients)


class MultipleRecipients(RecipientGenerator):
    """Recipient generator returning emails of entity with multiple recipients."""

    def __init__(self, key: str):
        """Initialize the generator."""
        self.key = key

    def __call__(self, notification: Notification, recipients: dict[str, Recipient]):
        """Generate recipients for the given entity."""
        fields = dict_lookup(notification.context, self.key)

        for idx, entity_dict in enumerate(fields):
            type_ = next(iter(entity_dict.keys()))
            key = f"{self.key}.{idx}.{type_}"
            generator = current_notification_recipient_generators_registry[type_](key, notification)
            generator(notification, recipients)


class GeneralRequestParticipantsRecipient(RecipientGenerator):
    """Generalization of invenio RequestParticipantsRecipient capable of working with general entities.

    RequestParticipantsRecipient supports only users.
    """

    def __init__(self, key: str):
        """Ctor."""
        self.key = key

    def __call__(self, notification: Notification, recipients: dict[str, Recipient]):
        """Fetch users involved in request and add as recipients."""
        request = dict_lookup(notification.context, self.key)

        EntityRecipientGenerator("request.created_by")(notification, recipients)
        EntityRecipientGenerator("request.receiver")(notification, recipients)

        # assume events can only be done by users
        # fetching all request events to get involved users
        request_events = current_events_service.scan(
            identity=system_identity,
            request_id=request["id"],
            extra_filter=dsl.Q("term", request_id=request["id"]),
        )
        # assume commenters can only be users
        user_ids = {re["created_by"]["user"] for re in request_events if re["created_by"].get("user")}

        filter_ = dsl.Q("terms", id=list(user_ids))
        users = current_users_service.scan(system_identity, extra_filter=filter_)
        for u in users:
            recipients[u["id"]] = Recipient(data=u)

        return recipients


# Do we need groups? Invenio doesn't seem to have GroupRecipient
"""
class GroupEmailRecipient(SpecificEntityRecipient):

    def _get_recipients(self, entity: Any) -> dict[str, Recipient]:
        return {
            user.email: Recipient(data=_extract_user_email_data(user))
            for user in entity.users.all()
            if getattr(user, "email", None)
        }
"""



#
# Copyright (c) 2025 CESNET z.s.p.o.
#
# This file is a part of oarepo-requests (see http://github.com/oarepo/oarepo-requests).
#
# oarepo-requests is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.
#
"""OARepo requests notifications module."""

from __future__ import annotations



#
# Copyright (C) 2025 CESNET z.s.p.o.
#
# oarepo-requests is free software; you can redistribute it and/or
# modify it under the terms of the MIT License; see LICENSE file for more
# details.
#
"""Default configuration for oarepo-requests to be initialized at invenio_config.module entrypoint."""

from __future__ import annotations

from oarepo_requests.resources.oarepo.config import OARepoRequestsResourceConfig
from oarepo_requests.resources.oarepo.resource import OARepoRequestsResource
from oarepo_requests.services.oarepo.config import OARepoRequestsServiceConfig
from oarepo_requests.services.oarepo.service import OARepoRequestsService

REQUESTS_SERVICE_CLASS = OARepoRequestsService
REQUESTS_SERVICE_CONFIG_CLASS = OARepoRequestsServiceConfig
REQUESTS_RESOURCE_CLASS = OARepoRequestsResource
REQUESTS_RESOURCE_CONFIG_CLASS = OARepoRequestsResourceConfig



#
# Copyright (C) 2024 CESNET z.s.p.o.
#
# oarepo-requests is free software; you can redistribute it and/or
# modify it under the terms of the MIT License; see LICENSE file for more
# details.
#
"""Publish draft request type."""

from __future__ import annotations

from typing import TYPE_CHECKING, Any, ClassVar, override

import marshmallow as ma
from invenio_i18n import gettext
from invenio_i18n import lazy_gettext as _

from .publish_base import PublishRequestType

if TYPE_CHECKING:
    from collections.abc import Mapping

    from flask_babel.speaklater import LazyString
    from flask_principal import Identity
    from invenio_records_resources.records import Record
    from invenio_requests.records.api import Request

    from ..utils import JsonValue


class PublishDraftRequestType(PublishRequestType):
    """Publish draft request type."""

    type_id = "publish_draft"
    name = _("Publish draft")

    payload_schema: Mapping[str, ma.fields.Field] | None = {
        "version": ma.fields.Str(),
    }

    @classmethod
    def is_applicable_to(cls, identity: Identity, topic: Record, *args: Any, **kwargs: Any) -> bool:
        """Check if the request type is applicable to the topic."""
        if cls.topic_type(topic) != "initial":
            return False

        return super().is_applicable_to(identity, topic, *args, **kwargs)

    # TODO: used in ui
    form: ClassVar[JsonValue] = {
        "field": "version",
        "ui_widget": "Input",
        "props": {
            "label": _("Resource version"),
            "placeholder": _("Write down the version (first, second…)."),
            "required": False,
        },
    }

    @override
    def stateful_name(
        self,
        identity: Identity,
        *,
        topic: Record,
        request: Request | None = None,
        **kwargs: Any,
    ) -> str | LazyString:
        """Return the stateful name of the request."""
        return self.string_by_state(
            identity=identity,
            topic=topic,
            request=request,
            create=gettext("Submit for review"),
            create_autoapproved=gettext("Publish draft"),
            submit=gettext("Submit for review"),
            submitted_receiver=gettext("Review and publish draft"),
            submitted_creator=gettext("Draft submitted for review"),
            submitted_others=gettext("Draft submitted for review"),
            accepted=gettext("Draft published"),
            declined=gettext("Draft publication declined"),
            cancelled=gettext("Draft publication cancelled"),
            created=gettext("Submit for review"),
        )

    @override
    def stateful_description(
        self,
        identity: Identity,
        *,
        topic: Record,
        request: Request | None = None,
        **kwargs: Any,
    ) -> str | LazyString:
        """Return the stateful description of the request."""
        return self.string_by_state(
            identity=identity,
            topic=topic,
            request=request,
            create=gettext(
                "By submitting the draft for review you are requesting the publication of the draft. "
                "The draft will become locked and no further changes will be possible until the request "
                "is accepted or declined. You will be notified about the decision by email."
            ),
            create_autoapproved=gettext(
                "Click to immediately publish the draft. "
                "The draft will be a subject to embargo as requested in the side panel. "
                "Note: The action is irreversible."
            ),
            submit=gettext(
                "Submit for review. After submitting the draft for review, "
                "it will be locked and no further modifications will be possible."
            ),
            submitted_receiver=gettext(
                "The draft has been submitted for review. You can now accept or decline the request."
            ),
            submitted_creator=gettext(
                "The draft has been submitted for review. "
                "It is now locked and no further changes are possible. "
                "You will be notified about the decision by email."
            ),
            submitted_others=gettext("The draft has been submitted for review. "),
            accepted=gettext("The draft has been published. "),
            declined=gettext("Publication of the draft has been declined."),
            cancelled=gettext("The draft has been cancelled. "),
            created=gettext("Waiting for finishing the draft publication request."),
        )



#
# Copyright (C) 2024 CESNET z.s.p.o.
#
# oarepo-requests is free software; you can redistribute it and/or
# modify it under the terms of the MIT License; see LICENSE file for more
# details.
#
"""Request type for requesting new version of a published record."""

from __future__ import annotations

from typing import TYPE_CHECKING, Any, ClassVar, override

import marshmallow as ma
from invenio_drafts_resources.records import Record as RecordWithDraft
from invenio_i18n import gettext
from invenio_i18n import lazy_gettext as _
from marshmallow.validate import OneOf
from oarepo_runtime.records.drafts import has_draft

from ..actions.new_version import NewVersionAcceptAction
from ..utils import classproperty, is_auto_approved, request_identity_matches
from .generic import NonDuplicableOARepoRecordRequestType

if TYPE_CHECKING:
    from collections.abc import Mapping

    from flask_babel.speaklater import LazyString
    from flask_principal import Identity
    from invenio_records_resources.records import Record
    from invenio_requests.customizations.actions import RequestAction
    from invenio_requests.records.api import Request

    from ..utils import JsonValue


class NewVersionRequestType(NonDuplicableOARepoRecordRequestType):
    """Request type for requesting new version of a published record."""

    type_id = "new_version"
    name = _("New Version")
    description = _("Request requesting creation of new version of a published record.")  # type: ignore[reportAssignmentType]
    allowed_on_draft = False
    editable = False
    payload_schema: Mapping[str, ma.fields.Field] | None = {
        "created_topic": ma.fields.Str(),
        "keep_files": ma.fields.String(validate=OneOf(["yes", "no"])),
    }

    @classproperty
    @override
    def available_actions(cls) -> dict[str, type[RequestAction]]:  # noqa N805 type: ignore[reportIncompatibleVariableOverride]
        """Return available actions for the request type."""
        return {
            **super().available_actions,
            "accept": NewVersionAcceptAction,
        }

    def can_create(
        self,
        identity: Identity,
        data: dict,
        receiver: dict[str, str],
        topic: Record,
        creator: dict[str, str],
        *args: Any,
        **kwargs: Any,
    ) -> None:
        """Check if the request can be created."""
        if not isinstance(topic, RecordWithDraft):
            raise TypeError(gettext("Trying to create edit request on record without draft support"))
        if has_draft(topic):
            raise ValueError(gettext("Trying to create edit request on record with draft"))
        super().can_create(identity, data, receiver, topic, creator, *args, **kwargs)

    @classmethod
    def is_applicable_to(cls, identity: Identity, topic: Record, *args: Any, **kwargs: Any) -> bool:
        """Check if the request type is applicable to the topic."""
        if not isinstance(topic, RecordWithDraft):
            raise TypeError(gettext("Trying to create edit request on record without draft support"))
        # if already editing metadata or a new version, we don't want to create a new request
        if has_draft(topic):
            return False
        return super().is_applicable_to(identity, topic, *args, **kwargs)

    # TODO: used in ui
    form: ClassVar[JsonValue] = {
        "field": "keep_files",
        "ui_widget": "Dropdown",
        "props": {
            "label": _("Keep files"),
            "placeholder": _("Yes or no"),
            "description": _(
                "If you choose yes, the current record's files will be linked to the new version of the record. "
                "Then you will be able to add/remove files in the form."
            ),
            "options": [
                {"id": "yes", "title_l10n": _("Yes")},
                {"id": "no", "title_l10n": _("No")},
            ],
        },
    }

    @override
    def stateful_name(
        self,
        identity: Identity,
        *,
        topic: Record,
        request: Request | None = None,
        **kwargs: Any,
    ) -> str | LazyString:
        """Return the stateful name of the request."""
        if is_auto_approved(self, identity=identity, topic=topic):
            return self.name
        if not request:
            return gettext("Request new version access")
        match request.status:
            case "submitted":
                return gettext("New version access requested")
            case _:
                return gettext("Request new version access")

    @override
    def stateful_description(  # noqa PLR0911
        self,
        identity: Identity,
        *,
        topic: Record,
        request: Request | None = None,
        **kwargs: Any,
    ) -> str | LazyString:
        """Return the stateful description of the request."""
        if is_auto_approved(self, identity=identity, topic=topic):
            return gettext("Click to start creating a new version of the record.")

        if not request:
            return gettext(
                "Request permission to update record (including files). "
                "You will be notified about the decision by email."
            )
        match request.status:
            case "submitted":
                if request_identity_matches(request.created_by, identity):
                    return gettext(
                        "Permission to update record (including files) requested. "
                        "You will be notified about the decision by email."
                    )
                if request_identity_matches(request.receiver, identity):
                    return gettext(
                        "You have been asked to approve the request to update the record. "
                        "You can approve or reject the request."
                    )
                return gettext("Permission to update record (including files) requested. ")
            case _:
                if request_identity_matches(request.created_by, identity):
                    return gettext("Submit request to get edit access to the record.")
                return gettext("You do not have permission to update the record.")



#
# Copyright (C) 2024 CESNET z.s.p.o.
#
# oarepo-requests is free software; you can redistribute it and/or
# modify it under the terms of the MIT License; see LICENSE file for more
# details.
#
"""Request types defined in oarepo-requests."""

from __future__ import annotations

from .delete_published_record import DeletePublishedRecordRequestType
from .edit_record import EditPublishedRecordRequestType
from .generic import NonDuplicableOARepoRecordRequestType
from .publish_draft import PublishDraftRequestType
from .ref_types import ModelRefTypes, ReceiverRefTypes

__all__ = [
    "DeletePublishedRecordRequestType",
    "EditPublishedRecordRequestType",
    "ModelRefTypes",
    "NonDuplicableOARepoRecordRequestType",
    "PublishDraftRequestType",
    "ReceiverRefTypes",
]



#
# Copyright (C) 2024 CESNET z.s.p.o.
#
# oarepo-requests is free software; you can redistribute it and/or
# modify it under the terms of the MIT License; see LICENSE file for more
# details.
#
"""Utility functions for the requests module."""

from __future__ import annotations

import copy
from typing import TYPE_CHECKING, Any, cast

from flask_babel.speaklater import LazyString
from invenio_access.permissions import system_identity
from invenio_drafts_resources.records import Record as RecordWithDraft
from invenio_drafts_resources.services import RecordService as DraftRecordService
from invenio_pidstore.errors import PersistentIdentifierError
from invenio_requests.proxies import current_request_type_registry
from invenio_requests.resolvers.registry import ResolverRegistry
from invenio_search.engine import dsl
from oarepo_runtime import current_runtime
from oarepo_workflows import (
    AutoApprove,
    Workflow,
    WorkflowRequest,
    WorkflowRequestPolicy,
)
from oarepo_workflows.errors import MissingWorkflowError
from oarepo_workflows.proxies import current_oarepo_workflows

from oarepo_requests.proxies import current_requests_service

if TYPE_CHECKING:
    from collections.abc import Callable, Mapping

    from flask_principal import Identity
    from invenio_records_resources.records import Record
    from invenio_records_resources.services import RecordService
    from invenio_requests.customizations.request_types import RequestType
    from invenio_requests.records.api import Request
    from invenio_requests.services.requests import RequestList
    from opensearch_dsl.query import Query

    from oarepo_requests.services.results import RequestTypesList

type JsonValue = str | LazyString | int | float | bool | None | dict[str, JsonValue] | list[JsonValue]


# TODO: move to runtime; typing issues
class classproperty[T]:  # noqa N801
    """Class property decorator as decorator chaining for declaring class properties was deprecated in python 3.11."""

    def __init__(self, func: Callable[[Any], T]):
        """Initialize the class property."""
        self.fget = func

    def __get__(self, instance: Any, owner: Any) -> T:
        """Get the value of the class property."""
        return self.fget(owner)


def allowed_request_types_for_record(identity: Identity, record: Record) -> dict[str, RequestType]:
    """Return allowed request types for the record.

    If there is a workflow defined on the record, only request types allowed by the workflow are returned.

    :param identity: Identity of the user. Only the request types for which user can create a request are returned.
    :param record: Record to get allowed request types for.
    :return: Dict of request types allowed for the record.
    """
    workflow_requests: WorkflowRequestPolicy | None
    try:
        workflow_requests = current_oarepo_workflows.get_workflow(record).requests()
        return {
            type_id: wr.request_type
            for (type_id, wr) in workflow_requests.applicable_workflow_requests(identity, record=record)
        }
    except MissingWorkflowError:
        # workflow not defined on the record, probably not a workflow-enabled record
        # so returning all matching request types
        pass

    record_ref = next(iter(reference_entity(record).keys()))

    ret = {}
    for rt in current_request_type_registry:
        if record_ref in rt.allowed_topic_ref_types:
            ret[rt.type_id] = rt

    return ret


def create_query_term_for_reference(field_name: str, reference: dict[str, str]) -> Query:
    """Create an opensearch query term for the reference.

    :param field_name: Field name to search in (can be "topic", "receiver", ...).
    :param reference: Reference to search for.
    :return: Opensearch query term.
    """
    return dsl.Q(
        "term",
        **{f"{field_name}.{next(iter(reference.keys()))}": next(iter(reference.values()))},
    )


def search_requests_filter(
    type_id: str,
    topic_reference: dict | None = None,
    receiver_reference: dict | None = None,
    creator_reference: dict | None = None,
    is_open: bool | None = None,
) -> Query:
    """Create a search filter for requests of a given request type.

    :param type_id: Request type id.
    :param topic_reference: Reference to the topic, optional
    :param receiver_reference: Reference to the receiver, optional
    :param creator_reference: Reference to the creator, optional
    :param is_open: Whether the request is open or closed. If not set, both open and closed requests are returned.
    """
    must = [
        dsl.Q("term", type=type_id),
    ]
    if is_open is not None:
        must.append(dsl.Q("term", is_open=is_open))
    if receiver_reference:
        must.append(create_query_term_for_reference("receiver", receiver_reference))
    if creator_reference:
        must.append(create_query_term_for_reference("creator", creator_reference))
    if topic_reference:
        must.append(create_query_term_for_reference("topic", topic_reference))

    return dsl.query.Bool(
        must=must,
    )


def open_request_exists(topic_or_reference: Record | dict[str, str], type_id: str) -> bool:
    """Check if there is an open request of a given type for the topic.

    :param topic_or_reference: Topic record or reference to the record in the form {"datasets": "id"}.
    :param type_id: Request type id.
    """
    topic_reference = ResolverRegistry.reference_entity(topic_or_reference, raise_=True)
    base_filter = search_requests_filter(type_id=type_id, topic_reference=topic_reference, is_open=True)
    results = current_requests_service.search(system_identity, extra_filter=base_filter).hits
    return bool(list(results))


def resolve_reference_dict(reference_dict: dict[str, str]) -> Any:
    """Resolve the reference dict to the entity (such as Record, User, ...).

    Raises ValueError if the reference cannot be resolved.
    """
    return ResolverRegistry.resolve_entity_proxy(reference_dict, raise_=True).resolve()  # type: ignore[reportOptionalMemberAccess]


def reference_entity(entity: Any) -> dict[str, str]:
    """Resolve the entity to the reference dict.

    Raises ValueError if the reference cannot be resolved.
    """
    return cast("dict[str, str]", ResolverRegistry.reference_entity(entity, raise_=True))


# TODO: possibly not used
def get_matching_service_for_refdict(
    reference_dict: dict[str, str],
) -> RecordService | None:
    """Get the service that is responsible for entities matching the reference dict.

    :param reference_dict: Reference dict in the form {"datasets": "id"}.
    :return: Service that is responsible for the entity or None if the entity does not have an associated service
    """
    for resolver in ResolverRegistry.get_registered_resolvers():
        if resolver.matches_reference_dict(reference_dict):
            return cast("RecordService", resolver.get_service())
    return None


def reference_to_tuple(reference: dict[str, str]) -> tuple[str, str]:
    """Convert the reference dict to a tuple.

    :param reference: Reference dict in the form {"datasets": "id"}.
    :return: Tuple in the form ("datasets", "id").
    """
    return next(iter(reference.items()))


def string_to_reference(reference_str: str) -> dict[str, str]:
    """Convert the reference string to a reference dict."""
    split = reference_str.split(":")
    return {split[0]: split[1]}


def ref_to_str(ref_dict: dict[str, str]) -> str:
    """Convert the reference string to a reference dict."""
    return f"{next(iter(ref_dict.keys()))}:{next(iter(ref_dict.values()))}"


def get_receiver_for_request_type(
    request_type: RequestType, identity: Identity, topic: Record
) -> Mapping[str, str] | None:
    """Get the default receiver for the request type, identity and topic.

    This call gets the workflow from the topic, looks up the request inside the workflow
    and evaluates workflow recipients for the request and topic and returns them.
    If the request has no matching receiver, None is returned.

    :param request_type: Request type.
    :param identity: Identity of the caller who wants to create a request of this type
    :param topic: Topic record for the request
    :return: Receiver for the request type from workflow or None if no receiver
    """
    if not topic:
        return None

    try:
        workflow: Workflow = current_oarepo_workflows.get_workflow(topic)
    except MissingWorkflowError:
        return None

    try:
        workflow_request: WorkflowRequest = workflow.requests()[request_type.type_id]
    except KeyError:
        return None

    return workflow_request.recipient_entity_reference(  # type: ignore[no-any-return]
        identity=identity, record=topic, request_type=request_type, creator=identity
    )


# TODO: consider moving to oarepo-workflows
def is_auto_approved(
    request_type: RequestType,
    *,
    identity: Identity,
    topic: Record,
) -> bool:
    """Check if the request should be auto-approved.

    If identity creates a request of the given request type on the given topic,
    the function checks if the request should be auto-approved.
    """
    if not current_oarepo_workflows:
        return False

    receiver = get_receiver_for_request_type(request_type=request_type, identity=identity, topic=topic)

    return bool(
        receiver
        and (isinstance(receiver, AutoApprove) or (isinstance(receiver, dict) and receiver.get("auto_approve")))
    )


def request_identity_matches(entity_reference: dict[str, str], identity: Identity) -> bool:
    """Check if the identity matches the entity reference.

    Identity matches the entity reference if the needs provided by the entity reference
    intersect with the needs provided by the identity. For example, if the entity reference
    provides [CommunityRoleNeed(comm_id, 'curator'), ActionNeed("administration")] and the
    identity provides [CommunityRoleNeed(comm_id, 'curator')], the function returns True.

    :param entity_reference: Entity reference in the form {"datasets": "id"}.
    :param identity:         Identity to check.
    """
    if not entity_reference:
        return False

    try:
        entity = ResolverRegistry.resolve_entity_proxy(entity_reference)
        if entity:
            needs = entity.get_needs()
            return bool(identity.provides.intersection(needs))
    except PersistentIdentifierError:
        return False
    return False


def merge_resource_configs[T](config_to_merge_in: T, original_config: Any) -> T:
    """Merge resource configurations."""
    actual_config = copy.deepcopy(config_to_merge_in)
    original_keys = {x for x in dir(original_config) if not x.startswith("_")}
    merge_in_keys = {
        x for x in dir(config_to_merge_in) if not x.startswith("_")
    }  # have to do this bc hasattr fails on resolving response_handlers
    for copy_from_original_key in original_keys - merge_in_keys:
        setattr(
            actual_config,
            copy_from_original_key,
            getattr(original_config, copy_from_original_key),
        )
    return actual_config


def has_rights_to_accept_request(request: Request, identity: Identity) -> bool:
    """Check if the identity has rights to accept the request.

    :param request: Request to check.
    :param identity: Identity to check.
    """
    return cast(
        "bool",
        current_requests_service.check_permission(
            identity,
            "action_accept",
            request=request,
            record=request.topic,
            request_type=request.type,
        ),
    )


def has_rights_to_submit_request(request: Request, identity: Identity) -> bool:
    """Check if the identity has rights to submit the request.

    :param request: Request to check.
    :param identity: Identity to check.
    """
    return cast(
        "bool",
        current_requests_service.check_permission(
            identity,
            "action_submit",
            request=request,
            record=request.topic,
            request_type=request.type,
        ),
    )


def search_requests(identity: Identity, record: RecordWithDraft | dict[str, str], expand: bool = False) -> RequestList:
    """Search requests for a given record."""
    topic_ref = reference_entity(record) if isinstance(record, RecordWithDraft) else record
    return cast(
        "RequestList",
        current_requests_service.search(identity, topic=topic_ref, expand=expand),
    )


def applicable_requests(identity: Identity, record: RecordWithDraft | dict[str, str]) -> RequestTypesList:
    """Get applicable request types for a record."""
    topic_ref = reference_entity(record) if isinstance(record, RecordWithDraft) else record
    return current_requests_service.applicable_request_types(identity, topic=topic_ref)


def get_draft_record_service(record: Record) -> DraftRecordService:
    """Get the draft record service for a record and checks it supports drafts."""
    topic_service = current_runtime.get_record_service_for_record(record)
    if not topic_service:
        raise KeyError(f"Record {record} service not found")
    if not isinstance(topic_service, DraftRecordService):
        raise TypeError("Draft service required for editing records.")
    return topic_service



#
# Copyright (C) 2024 CESNET z.s.p.o.
#
# oarepo-requests is free software; you can redistribute it and/or
# modify it under the terms of the MIT License; see LICENSE file for more
# details.
#
"""Publish draft request type."""

from __future__ import annotations

from typing import TYPE_CHECKING, Any, override

from invenio_i18n import gettext
from invenio_i18n import lazy_gettext as _

from .publish_base import PublishRequestType

if TYPE_CHECKING:
    from flask_babel.speaklater import LazyString
    from flask_principal import Identity
    from invenio_records_resources.records import Record
    from invenio_requests.records.api import Request


class PublishChangedMetadataRequestType(PublishRequestType):
    """Request type for publication of changed metadata."""

    type_id = "publish_changed_metadata"
    name = _("Publish changed metadata")

    @classmethod
    def is_applicable_to(cls, identity: Identity, topic: Record, *args: Any, **kwargs: Any) -> bool:
        """Check if the request type is applicable to the topic."""
        if cls.topic_type(topic) != "metadata":
            return False

        return super().is_applicable_to(identity, topic, *args, **kwargs)

    # TODO: used in ui
    @override
    def stateful_name(
        self,
        identity: Identity,
        *,
        topic: Record,
        request: Request | None = None,
        **kwargs: Any,
    ) -> str | LazyString:
        """Return the stateful name of the request."""
        return self.string_by_state(
            identity=identity,
            topic=topic,
            request=request,
            create=gettext("Submit for review"),
            create_autoapproved=gettext("Publish changed metadata"),
            submit=gettext("Submit for review"),
            submitted_receiver=gettext("Review and publish changed metadata"),
            submitted_creator=gettext("Changed metadata submitted for review"),
            submitted_others=gettext("Changed metadata submitted for review"),
            accepted=gettext("Accepted changed metadata publication"),
            declined=gettext("Declined changed metadata publication"),
            cancelled=gettext("Cancelled changed metadata publication"),
            created=gettext("Submit for review"),
        )

    @override
    def stateful_description(
        self,
        identity: Identity,
        *,
        topic: Record,
        request: Request | None = None,
        **kwargs: Any,
    ) -> str | LazyString:
        """Return the stateful description of the request."""
        return self.string_by_state(
            identity=identity,
            topic=topic,
            request=request,
            create=gettext(
                "By submitting the changed metadata for review you are requesting the publication of the "
                "changed metadata. "
                "The draft will become locked and no further changes will be possible until the request "
                "is accepted or declined. You will be notified about the decision by email."
            ),
            create_autoapproved=gettext(
                "Click to immediately publish the changed metadata. "
                "The draft will be a subject to embargo as requested in the side panel. "
                "Note: The action is irreversible."
            ),
            submit=gettext(
                "Submit for review. After submitting the changed metadata for review, "
                "it will be locked and no further modifications will be possible."
            ),
            submitted_receiver=gettext(
                "The record with changed metadata has been submitted for review. "
                "You can now accept or decline the request."
            ),
            submitted_creator=gettext(
                "The record with changed metadata has been submitted for review. "
                "It is now locked and no further changes are possible. "
                "You will be notified about the decision by email."
            ),
            submitted_others=gettext("The record with changed metadata has been submitted for review. "),
            accepted=gettext("Accepted changed metadata publication"),
            declined=gettext("Declined changed metadata publication"),
            cancelled=gettext("Cancelled changed metadata publication"),
            created=gettext("Waiting for finishing the changed metadata publication request."),
        )



#
# Copyright (C) 2024 CESNET z.s.p.o.
#
# oarepo-requests is free software; you can redistribute it and/or
# modify it under the terms of the MIT License; see LICENSE file for more
# details.
#
"""Actions for delete published record request."""

from __future__ import annotations

from typing import TYPE_CHECKING, Any, cast, override

from invenio_notifications.services.uow import NotificationOp

from ..notifications.builders.delete_published_record import (
    DeletePublishedRecordRequestAcceptNotificationBuilder,
    DeletePublishedRecordRequestDeclineNotificationBuilder,
    DeletePublishedRecordRequestSubmitNotificationBuilder,
)
from .generic import OARepoAcceptAction, OARepoDeclineAction, OARepoSubmitAction

if TYPE_CHECKING:
    from flask_principal import Identity

from typing import TYPE_CHECKING

from invenio_db import db
from invenio_i18n import _
from oarepo_runtime.proxies import current_runtime
from oarepo_runtime.typing import record_from_result

if TYPE_CHECKING:
    from flask_principal import Identity


if TYPE_CHECKING:
    from flask_principal import Identity
    from invenio_db.uow import UnitOfWork
    from invenio_rdm_records.services.services import RDMRecordService


class DeletePublishedRecordSubmitAction(OARepoSubmitAction):
    """Submit the delete published record request."""

    @override
    def apply(
        self,
        identity: Identity,
        uow: UnitOfWork,
        *args: Any,
        **kwargs: Any,
    ) -> None:
        uow.register(NotificationOp(DeletePublishedRecordRequestSubmitNotificationBuilder.build(request=self.request)))


class DeletePublishedRecordAcceptAction(OARepoAcceptAction):
    """Accept request for deletion of a published record and delete the record."""

    name = _("Permanently delete")

    @override
    def apply(
        self,
        identity: Identity,
        uow: UnitOfWork,
        *args: Any,
        **kwargs: Any,
    ) -> None:
        topic = self.topic
        topic_service = current_runtime.get_record_service_for_record(topic)
        if not topic_service:
            raise KeyError(f"topic {topic} service not found")
        if hasattr(topic_service, "delete_record"):
            topic_service = cast("RDMRecordService", topic_service)
            from flask import current_app

            oarepo = current_app.extensions["oarepo-runtime"]
            resource_config = oarepo.models_by_record_class[topic_service.record_cls].resource_config

            citation_text = "Citation unavailable."
            if resource_config and "text/x-iso-690+plain" in resource_config.response_handlers:
                citation_text = resource_config.response_handlers["text/x-iso-690+plain"].serializer.serialize_object(
                    topic
                )

            data = {
                "removal_reason": {"id": self.request["payload"]["removal_reason"]},
                "citation_text": citation_text,
                "note": self.request["payload"].get("note", ""),
                "is_visible": True,
            }
            self.topic = record_from_result(topic_service.delete_record(identity, topic["id"], data))
            db.session.commit()
        else:
            topic_service.delete(identity, topic["id"], *args, uow=uow, **kwargs)

        uow.register(NotificationOp(DeletePublishedRecordRequestAcceptNotificationBuilder.build(request=self.request)))
        # TODO: cascade cancel?


class DeletePublishedRecordDeclineAction(OARepoDeclineAction):
    """Decline request for deletion of a published record."""

    name = _("Keep the record")

    @override
    def apply(
        self,
        identity: Identity,
        uow: UnitOfWork,
        *args: Any,
        **kwargs: Any,
    ) -> None:
        uow.register(NotificationOp(DeletePublishedRecordRequestDeclineNotificationBuilder.build(request=self.request)))



#
# Copyright (c) 2025 CESNET z.s.p.o.
#
# This file is a part of oarepo-requests (see http://github.com/oarepo/oarepo-requests).
#
# oarepo-requests is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.
#
"""OARepo notification builders for delete publish request notifications."""

from __future__ import annotations

from ..generators import EntityRecipientGenerator, ReferenceSavingEntityResolve
from .base import RequestActionNotificationBuilder


class DeletePublishedRecordRequestSubmitNotificationBuilder(RequestActionNotificationBuilder):
    """Notification builder for delete published record request submit event."""

    type = "delete-published-record-request-event.submit"

    recipients = (EntityRecipientGenerator(key="request.receiver"),)  # email only


class DeletePublishedRecordRequestAcceptNotificationBuilder(RequestActionNotificationBuilder):
    """Notification builder for delete published record request accept event."""

    type = "delete-published-record-request-event.accept"

    recipients = (EntityRecipientGenerator(key="request.created_by"),)

    # topic resolution crashes on RecordDeletedException
    context = (
        ReferenceSavingEntityResolve(key="request"),
        ReferenceSavingEntityResolve(key="request.created_by"),
    )


class DeletePublishedRecordRequestDeclineNotificationBuilder(RequestActionNotificationBuilder):
    """Notification builder for delete published record request decline event."""

    type = "delete-published-record-request-event.decline"

    recipients = (EntityRecipientGenerator(key="request.created_by"),)



#
# Copyright (C) 2024 CESNET z.s.p.o.
#
# oarepo-requests is free software; you can redistribute it and/or
# modify it under the terms of the MIT License; see LICENSE file for more
# details.
#
"""Actions for publishing draft requests."""

from __future__ import annotations

from typing import TYPE_CHECKING, Any, override

from invenio_access.permissions import system_identity
from invenio_i18n import _
from invenio_notifications.services.uow import NotificationOp
from invenio_requests.records.api import Request
from oarepo_runtime.typing import record_from_result

from oarepo_requests.errors import UnresolvedRequestsError, VersionAlreadyExists

from ..notifications.builders.publish import (
    PublishDraftRequestAcceptNotificationBuilder,
    PublishDraftRequestDeclineNotificationBuilder,
    PublishDraftRequestSubmitNotificationBuilder,
)
from ..utils import get_draft_record_service, search_requests
from .generic import (
    OARepoAcceptAction,
    OARepoDeclineAction,
    OARepoSubmitAction,
)

if TYPE_CHECKING:
    from flask_principal import Identity
    from invenio_db.uow import UnitOfWork
    from invenio_requests.customizations import RequestAction
else:
    RequestAction = object


class PublishMixin(RequestAction):
    """Mixin for publish actions."""

    def can_execute(self) -> bool:
        """Check if the action can be executed."""
        if not super().can_execute():
            return False

        try:
            from ..types.publish_draft import PublishDraftRequestType

            topic = self.request.topic.resolve()
            PublishDraftRequestType.validate_topic(system_identity, topic)
        except:  # noqa E722: used for displaying buttons, so ignore errors here
            return False
        return True


# TODO: snapshot
class PublishDraftSubmitAction(PublishMixin, OARepoSubmitAction):
    """Submit action for publishing draft requests."""

    @override
    def apply(
        self,
        identity: Identity,
        uow: UnitOfWork,
        *args: Any,
        **kwargs: Any,
    ) -> None:
        if "payload" in self.request and "version" in self.request["payload"]:
            topic_service = get_draft_record_service(self.topic)
            versions = topic_service.search_versions(identity, self.topic.pid.pid_value)
            versions_hits = versions.to_dict()["hits"]["hits"]
            for rec in versions_hits:
                if "version" in rec["metadata"]:
                    version = rec["metadata"]["version"]
                    if version == self.request["payload"]["version"]:
                        raise VersionAlreadyExists
            self.topic.metadata["version"] = self.request["payload"]["version"]
        uow.register(NotificationOp(PublishDraftRequestSubmitNotificationBuilder.build(request=self.request)))


class PublishDraftAcceptAction(PublishMixin, OARepoAcceptAction):
    """Accept action for publishing draft requests."""

    name = _("Publish")

    @override
    def apply(
        self,
        identity: Identity,
        uow: UnitOfWork,
        *args: Any,
        **kwargs: Any,
    ) -> None:
        topic_service = get_draft_record_service(self.topic)
        requests = search_requests(system_identity, self.topic)

        for result in requests._results:  # noqa SLF001
            if (
                result.type
                not in [
                    "publish_draft",
                    "publish_new_version",
                    "publish_changed_metadata",
                ]
                and result.is_open
                and Request.get_record(result.uuid)["status"]
                in (
                    "submitted",
                    "created",
                )
            ):
                # note: we can not use solely the result.is_open because changes may not be committed yet
                # to opensearch index. That's why we need to get the record from DB and re-check.
                raise UnresolvedRequestsError(action=str(self.name))
        id_ = self.topic["id"]
        self.topic = record_from_result(topic_service.publish(identity, id_, *args, uow=uow, expand=False, **kwargs))
        uow.register(NotificationOp(PublishDraftRequestAcceptNotificationBuilder.build(request=self.request)))


class PublishDraftDeclineAction(OARepoDeclineAction):
    """Decline action for publishing draft requests."""

    name = _("Return for correction")

    @override
    def apply(
        self,
        identity: Identity,
        uow: UnitOfWork,
        *args: Any,
        **kwargs: Any,
    ) -> None:
        """Publish the draft."""
        uow.register(NotificationOp(PublishDraftRequestDeclineNotificationBuilder.build(request=self.request)))



#
# Copyright (C) 2024 CESNET z.s.p.o.
#
# oarepo-requests is free software; you can redistribute it and/or
# modify it under the terms of the MIT License; see LICENSE file for more
# details.
#
"""Actions for creating a new version of a published record."""

from __future__ import annotations

from typing import TYPE_CHECKING, Any, override

from oarepo_runtime.typing import record_from_result

from ..utils import get_draft_record_service, ref_to_str, reference_entity
from .generic import OARepoAcceptAction

if TYPE_CHECKING:
    from flask_principal import Identity
    from invenio_db.uow import UnitOfWork


# TODO: snapshot
class NewVersionAcceptAction(OARepoAcceptAction):
    """Accept creation of a new version of a published record."""

    @override
    def apply(
        self,
        identity: Identity,
        uow: UnitOfWork,
        *args: Any,
        **kwargs: Any,
    ) -> None:
        """Apply the action, creating a new version of the record."""
        topic_service = get_draft_record_service(self.topic)
        new_version_topic = topic_service.new_version(identity, self.topic["id"], uow=uow)
        created_topic = record_from_result(new_version_topic)
        if (
            "payload" in self.request
            and "keep_files" in self.request["payload"]
            and self.request["payload"]["keep_files"] == "yes"
        ):
            topic_service.import_files(identity, new_version_topic.id)
        if "payload" not in self.request:
            self.request["payload"] = {}
        # payload only allows strings
        self.request["payload"]["created_topic"] = ref_to_str(reference_entity(created_topic))



#
# Copyright (C) 2024 CESNET z.s.p.o.
#
# oarepo-requests is free software; you can redistribute it and/or
# modify it under the terms of the MIT License; see LICENSE file for more
# details.
#
"""Permissions for requests based on workflows."""

from __future__ import annotations

from oarepo_workflows.requests.permissions import (
    CreatorsFromWorkflowRequestsPermissionPolicy,
)
from oarepo_workflows.services.permissions import DefaultWorkflowPermissions

from oarepo_requests.services.permissions.generators.active import RequestActive


class RequestBasedWorkflowPermissions(DefaultWorkflowPermissions):
    """Base class for workflow permissions, subclass from it and put the result to Workflow constructor.

    This permission adds a special generator RequestActive() to the default permissions.
    Whenever the request is in `accept` action, the RequestActive generator matches.

    Example:
        class MyWorkflowPermissions(RequestBasedWorkflowPermissions):
            can_read = [AnyUser()]
    in invenio.cfg
    WORKFLOWS = {
        'default': Workflow(
            permission_policy_cls = MyWorkflowPermissions, ...
        )
    }

    """

    can_delete = (*DefaultWorkflowPermissions.can_delete, RequestActive())
    can_publish = (RequestActive(),)
    can_edit = (RequestActive(),)
    can_new_version = (RequestActive(),)


__all__ = (
    "CreatorsFromWorkflowRequestsPermissionPolicy",
    "RequestBasedWorkflowPermissions",
)



#
# Copyright (c) 2025 CESNET z.s.p.o.
#
# This file is a part of oarepo-requests (see https://github.com/oarepo/oarepo-communities).
#
# oarepo-communities is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.
#
"""Filters for notifications."""

from __future__ import annotations

from typing import TYPE_CHECKING

from invenio_notifications.services.filters import RecipientFilter

if TYPE_CHECKING:
    from invenio_notifications.models import Notification, Recipient


class UsersWithNoMailRecipientFilter(RecipientFilter):
    """Recipient filter for filtering system user."""

    def __call__(self, notification: Notification, recipients: dict[str, Recipient]):  # noqa ARG002
        """Filter system recipient."""
        return {id_: recipient for id_, recipient in recipients.items() if "email" in recipient.data}


class SystemUserRecipientFilter(RecipientFilter):
    """Recipient filter for filtering system user."""

    def __call__(self, notification: Notification, recipients: dict[str, Recipient]):  # noqa ARG002
        """Filter system recipient."""
        recipients.pop("system", None)
        return recipients



#
# Copyright (c) 2025 CESNET z.s.p.o.
#
# This file is a part of oarepo-requests (see http://github.com/oarepo/oarepo-requests).
#
# oarepo-requests is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.
#
"""OARepo notification builders for comment event notification."""

from __future__ import annotations

from typing import TYPE_CHECKING

from invenio_notifications.services.generators import (
    ContextGenerator,
    EntityResolve,
    RecipientGenerator,
)
from invenio_requests.notifications.builders import (
    CommentRequestEventCreateNotificationBuilder as InvenioCommentRequestEventCreateNotificationBuilder,
)
from invenio_requests.notifications.generators import RequestParticipantsRecipient

from oarepo_requests.notifications.filters import UsersWithNoMailRecipientFilter
from oarepo_requests.notifications.generators import (
    GeneralRequestParticipantsRecipient,
    ReferenceSavingEntityResolve,
    RequestEntityResolve,
)
from oarepo_requests.utils import classproperty

if TYPE_CHECKING:
    from invenio_notifications.services.filters import RecipientFilter


class CommentRequestEventCreateNotificationBuilder(InvenioCommentRequestEventCreateNotificationBuilder):
    """Notification builder for comment event."""

    @classproperty
    def context(self) -> tuple[ContextGenerator, ...]:  # type: ignore[reportIncompatibleVariableOverride]
        """Get context for the notification builder."""
        invenio = super().context
        for idx, r in enumerate(list(invenio)):
            if isinstance(r, EntityResolve):
                replacement_cls = RequestEntityResolve if r.key == "request" else ReferenceSavingEntityResolve
                invenio[idx] = replacement_cls(key=r.key)
        return *invenio, ReferenceSavingEntityResolve(key="request.topic")

    @classproperty
    def recipients(self) -> tuple[RecipientGenerator, ...]:  # type: ignore[reportIncompatibleVariableOverride]
        """Get recipients for the notification builder."""
        invenio = super().recipients
        for idx, r in enumerate(list(invenio)):
            if isinstance(r, RequestParticipantsRecipient):
                invenio[idx] = GeneralRequestParticipantsRecipient(key="request")
                break
        return tuple(invenio)

    @classproperty
    def recipient_filters(self) -> tuple[RecipientFilter, ...]:  # type: ignore[reportIncompatibleVariableOverride]
        """Get recipient filters for the notification builder."""
        return *super().recipient_filters, UsersWithNoMailRecipientFilter()



#
# Copyright (C) 2024 CESNET z.s.p.o.
#
# oarepo-requests is free software; you can redistribute it and/or
# modify it under the terms of the MIT License; see LICENSE file for more
# details.
#
"""Classes that define allowed reference types for the topic and receiver references."""

from __future__ import annotations

from typing import Self

from oarepo_runtime import current_runtime

from oarepo_requests.proxies import current_oarepo_requests


class ModelRefTypes:
    """Class is used to define the allowed reference types for the topic reference."""

    def __get__(self, obj: Self, owner: type[Self]) -> list[str]:
        """Property getter, returns the list of allowed reference types."""
        return [model.entity_type for model in current_runtime.rdm_models]


class ReceiverRefTypes:
    """Class is used to define the allowed reference types for the receiver reference.

    The list of ref types is taken from the configuration (configuration key REQUESTS_ALLOWED_RECEIVERS).
    """

    def __get__(self, obj: Self, owner: type[Self]) -> list[str]:
        """Property getter, returns the list of allowed reference types."""
        return current_oarepo_requests.allowed_receiver_ref_types



#
# Copyright (C) 2025 CESNET z.s.p.o.
#
# oarepo-requests is free software; you can redistribute it and/or
# modify it under the terms of the MIT License; see LICENSE file for more
# details.
#
"""Notification generators in oarepo-requests."""

from __future__ import annotations

from oarepo_requests.notifications.generators.context import (
    ReferenceSavingEntityResolve,
    RequestEntityResolve,
)
from oarepo_requests.notifications.generators.recipients import (
    EntityRecipientGenerator,
    GeneralRequestParticipantsRecipient,
    MultipleRecipients,
)

__all__ = [
    "EntityRecipientGenerator",
    "GeneralRequestParticipantsRecipient",
    "MultipleRecipients",
    "ReferenceSavingEntityResolve",
    "RequestEntityResolve",
]



#
# Copyright (C) 2024 CESNET z.s.p.o.
#
# oarepo-requests is free software; you can redistribute it and/or
# modify it under the terms of the MIT License; see LICENSE file for more
# details.
#
"""Actions for creating a draft of published record for editing metadata."""

from __future__ import annotations

from typing import TYPE_CHECKING, Any, override

from oarepo_runtime.typing import record_from_result

from ..utils import get_draft_record_service
from .generic import OARepoAcceptAction

if TYPE_CHECKING:
    from flask_principal import Identity
    from invenio_db.uow import UnitOfWork


# TODO: snapshot
class EditTopicAcceptAction(OARepoAcceptAction):
    """Accept creation of a draft of a published record for editing metadata."""

    @override
    def apply(
        self,
        identity: Identity,
        uow: UnitOfWork,
        *args: Any,
        **kwargs: Any,
    ) -> None:
        topic_service = get_draft_record_service(self.topic)
        self.topic = record_from_result(topic_service.edit(identity, self.topic["id"], uow=uow))



#
# Copyright (c) 2025 CESNET z.s.p.o.
#
# This file is a part of oarepo-requests (see http://github.com/oarepo/oarepo-requests).
#
# oarepo-requests is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.
#
"""OARepo notification builders for publish draft notifications."""

from __future__ import annotations

from ..generators import EntityRecipientGenerator
from .base import RequestActionNotificationBuilder


class PublishDraftRequestSubmitNotificationBuilder(RequestActionNotificationBuilder):
    """Notification builder for publish draft request submit event."""

    type = "publish-draft-request-event.submit"

    recipients = (EntityRecipientGenerator(key="request.receiver"),)  # email only


class PublishDraftRequestAcceptNotificationBuilder(RequestActionNotificationBuilder):
    """Notification builder for publish draft request accept event."""

    type = "publish-draft-request-event.accept"

    recipients = (EntityRecipientGenerator(key="request.created_by"),)


class PublishDraftRequestDeclineNotificationBuilder(RequestActionNotificationBuilder):
    """Notification builder for publish draft request decline event."""

    type = "publish-draft-request-event.decline"

    recipients = (EntityRecipientGenerator(key="request.created_by"),)



#
# Copyright (C) 2025 CESNET z.s.p.o.
#
# oarepo-requests is free software; you can redistribute it and/or
# modify it under the terms of the MIT License; see LICENSE file for more
# details.
#
#
# This file is part of Invenio.
# Copyright (C) 2016-2018 CERN.
#
# Invenio is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Create snapshot table."""

from __future__ import annotations

import sqlalchemy as sa
import sqlalchemy_utils
from alembic import op
from sqlalchemy.dialects import postgresql

# revision identifiers, used by Alembic.
revision = "57aebe908bcb"
down_revision = "985368490c5b"
branch_labels = ()
depends_on = None


def upgrade() -> None:
    """Upgrade database."""
    # ### commands auto generated by Alembic - please adjust! ###
    op.create_table(
        "record_snapshots",
        sa.Column("id", sa.Integer(), nullable=False),
        sa.Column("record_uuid", sqlalchemy_utils.types.uuid.UUIDType(), nullable=False),
        sa.Column(
            "json",
            sa.JSON()
            .with_variant(sqlalchemy_utils.types.json.JSONType(), "mysql")
            .with_variant(postgresql.JSONB(none_as_null=True, astext_type=sa.Text()), "postgresql")
            .with_variant(sqlalchemy_utils.types.json.JSONType(), "sqlite"),
            nullable=False,
        ),
        sa.Column("request_id", sqlalchemy_utils.types.uuid.UUIDType(), nullable=True),
        sa.Column("created", sa.DateTime(), nullable=False),
        sa.ForeignKeyConstraint(
            ["request_id"],
            ["request_metadata.id"],
            name=op.f("fk_record_snapshots_request_id_request_metadata"),
            ondelete="cascade",
        ),
        sa.PrimaryKeyConstraint("id", name=op.f("pk_record_snapshots")),
    )
    # ### end Alembic commands ###


def downgrade() -> None:
    """Downgrade database."""
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_table("record_snapshots")
    # ### end Alembic commands ###



# noqa INP001

#
# Copyright (C) 2024 CESNET z.s.p.o.
#
# oarepo-requests is free software; you can redistribute it and/or
# modify it under the terms of the MIT License; see LICENSE file for more
# details.
#
from __future__ import annotations

from invenio_i18n import _

_("Create Request")
_("Open dialog for request")
_("My Requests")
_("Requests to Approve")
_("Are you sure?")
_("Cancel")
_("OK")
_("Create request")
_("Submit request")
_("Delete request")
_("Delete")
_("Accept request")
_("Accept")
_("Decline request")
_("Decline")
_("Create and submit request")
_("Create and submit")
_("Error sending request")
_("Submit")
_("Save drafted request")
_("Save")
_("Create")
_("Creator")
_("Receiver")
_("Request type")
_("Created")
_("Timeline")
_("Submit event")
_("No requests to show")
_("api.requests")
_("api.applicable-requests")



#
# Copyright (C) 2024 CESNET z.s.p.o.
#
# oarepo-requests is free software; you can redistribute it and/or
# modify it under the terms of the MIT License; see LICENSE file for more
# details.
#
"""Config for the extended requests API."""

from __future__ import annotations

from types import MappingProxyType
from typing import TYPE_CHECKING

import importlib_metadata
import marshmallow as ma
from invenio_records_resources.services.base.config import ConfiguratorMixin
from invenio_requests.resources import RequestsResourceConfig

from oarepo_requests.services.search import ExtendedRequestSearchRequestArgsSchema

if TYPE_CHECKING:
    from collections.abc import Callable, Mapping
    from http.client import HTTPException

    from flask import Response
    from flask_resources import ResponseHandler
    from marshmallow import fields


class OARepoRequestsResourceConfig(RequestsResourceConfig, ConfiguratorMixin):
    """Config for the oarepo requests API."""

    routes: Mapping[str, str] = {
        **RequestsResourceConfig.routes,
        "create": "/<topic>/<request_type>",
        "list-applicable": "/applicable",
    }

    request_search_args = ExtendedRequestSearchRequestArgsSchema

    @property
    def request_view_args(self) -> Mapping[str, fields.Field]:  # type: ignore[reportIncompatibleVariableOverride]
        """Request view args for the requests API."""
        return {
            **super().request_view_args,
            "topic": ma.fields.String(),
            "request_type": ma.fields.String(),
        }

    @property
    def response_handlers(self) -> Mapping[str, ResponseHandler]:  # type: ignore[reportIncompatibleVariableOverride]
        """Response handlers for the extended requests API."""
        return {
            # TODO: UI serialization
            **super().response_handlers,
        }

    @property
    def error_handlers(  # type: ignore[reportIncompatibleVariableOverride]
        self,
    ) -> Mapping[
        type[HTTPException | BaseException] | int,
        Callable[[Exception], Response],
    ]:
        """Get error handlers."""
        error_handlers = dict(super().error_handlers)
        for x in importlib_metadata.entry_points(group="oarepo_requests.error_handlers"):
            error_handlers.update(x.load())
        return MappingProxyType(error_handlers)  # type: ignore[reportReturnType]
